"""This module contains user interface related code, such as the main screen
and diagram windows."""

# ruff: noqa: E402

from __future__ import annotations

import logging
import os
import sys

import gi

gi.require_version("Gtk", "4.0")
gi.require_version("Gdk", "4.0")
gi.require_version("GtkSource", "5")
gi.require_version("Adw", "1")

from gi.repository import Adw, Gio, GLib, Gtk, GtkSource

import gaphor.asyncio
import gaphor.ui.diagramview  # noqa: F401
import gaphor.ui.textfield  # noqa: F401
from gaphor.application import Application, Session
from gaphor.core import event_handler
from gaphor.event import ActiveSessionChanged, ApplicationShutdown, SessionCreated
from gaphor.i18n import gettext, translated_ui_string
from gaphor.settings import APPLICATION_ID, StyleVariant, settings
from gaphor.storage.recovery import all_sessions
from gaphor.ui.actiongroup import (
    apply_application_actions,
    apply_shortcuts_from_entry_point,
)

Adw.init()
GtkSource.init()

log = logging.getLogger(__name__)


def run(argv: list[str], *, launch_service="greeter", recover=False) -> int:  # noqa: C901
    application: Application | None = None

    def app_handle_local_options(gtk_app, options):
        if options.contains("greeter"):
            gtk_app.register()
            if gtk_app.get_property("is-remote"):
                gtk_app.activate_action("new-window")
                return 0

        return -1

    def app_startup(gtk_app):
        nonlocal application

        @event_handler(SessionCreated)
        def on_session_created(event: SessionCreated):
            event_manager = event.session.get_service("event_manager")
            event_manager.subscribe(on_session_changed)
            main_window = event.session.get_service("main_window")
            main_window.open(gtk_app)

        @event_handler(ActiveSessionChanged)
        def on_session_changed(event: ActiveSessionChanged):
            if isinstance(event.service, Session):
                main_window = event.service.get_service("main_window")
                if main_window.window:
                    main_window.window.present()

        @event_handler(ApplicationShutdown)
        def on_quit(_event: ApplicationShutdown):
            gtk_app.quit()

        apply_shortcuts_from_entry_point(
            "gaphor.appservices", "app", gtk_app, extra=[Application]
        )
        apply_shortcuts_from_entry_point("gaphor.services", "win", gtk_app)
        apply_shortcuts_from_entry_point("gaphor.services", "text", gtk_app)

        try:
            application = Application(gtk_app=gtk_app)
            apply_application_actions(application, gtk_app)

            event_manager = application.get_service("event_manager")
            event_manager.subscribe(on_session_created)
            event_manager.subscribe(on_quit)
            application.get_service(launch_service).init(gtk_app)
            if recover:
                recover_sessions(application)
        except Exception:
            gtk_app.exit_code = 1
            gtk_app.quit()
            raise

    def app_activate(gtk_app):
        assert application
        if not application.sessions:
            application.get_service(launch_service).open()
        elif gtk_app.get_active_window():
            gtk_app.get_active_window().present()

    def app_open(gtk_app, files, n_files, hint):
        # appfilemanager should take care of this:
        assert application
        if hint == "__new__":
            application.new_session()
        else:
            for file in files:
                application.new_session(filename=file.get_path())

    def update_color_scheme(style_variant: StyleVariant):
        gtk_app.get_style_manager().set_color_scheme(
            {
                StyleVariant.DARK: Adw.ColorScheme.FORCE_DARK,
                StyleVariant.LIGHT: Adw.ColorScheme.FORCE_LIGHT,
            }.get(style_variant, Adw.ColorScheme.DEFAULT)
        )

    gtk_app = UIApplication()

    settings.style_variant_changed(update_color_scheme)
    gtk_app.exit_code = 0

    gtk_app.add_main_option(
        "greeter",
        b"g",
        GLib.OptionFlags.NONE,
        GLib.OptionArg.NONE,
        "Create a new model",
        None,
    )

    gtk_app.connect("handle-local-options", app_handle_local_options)
    gtk_app.connect("startup", app_startup)
    gtk_app.connect("activate", app_activate)
    gtk_app.connect("open", app_open)

    with gaphor.asyncio.glib_event_loop_policy():
        gtk_app.run(argv)

    return gtk_app.exit_code


def recover_sessions(application):
    for session_id, filename, template in all_sessions():
        if not any(
            session.session_id == session_id for session in application.sessions
        ):
            application.recover_session(
                session_id=session_id, filename=filename, template=template
            )


def macos_menubar():
    if sys.platform != "darwin":
        return False

    supported = (Gtk.MAJOR_VERSION, Gtk.MINOR_VERSION) >= (4, 17)
    flags = os.getenv("GAPHOR_FEATURE_FLAG", "").split(",")
    return supported and "nomenubar" not in flags


class UIApplication(Adw.Application):
    def __init__(self):
        # Register session on Darwin, so the NSApplicationDelegate is registered for opening files
        super().__init__(
            application_id=APPLICATION_ID,
            flags=Gio.ApplicationFlags.HANDLES_OPEN,
            register_session=(sys.platform == "darwin"),
        )
        self._menus = {}

    def do_startup(self):
        Adw.Application.do_startup(self)

        builder = Gtk.Builder()
        builder.add_from_string(translated_ui_string("gaphor.ui", "menubar.ui"))
        if macos_menubar():
            self.set_menubar(builder.get_object("menu"))
            # Set keyboard shortcuts on toplevel, so they appear in the menu
            self.set_accels_for_action("clipboard.paste-full", ["<Meta><Shift>v"])
            self.set_accels_for_action(
                "selection.delete", ["Delete", "BackSpace", "<Meta>BackSpace"]
            )
            self.set_accels_for_action("selection.unselect-all", ["<Meta><Shift>a"])

        self._menus["export"] = (gettext("Export"), builder.get_object("export-menu"))
        self._menus["tools"] = (gettext("Tools"), builder.get_object("tools-menu"))

    def update_menu(self, name, submenu):
        hook = self._menus.get(name)
        if hook:
            title, menu = hook
            menu.remove_all()
            menu.append_submenu(title, submenu)
        else:
            log.warning("No submenu with name %s", name)


if __name__ == "__main__":
    run(sys.argv)



from __future__ import annotations

from gaphor.diagram.presentation import ElementPresentation


def align_left(elements: set[ElementPresentation]):
    left_edge = _left_edge(elements)

    for item in elements:
        item.matrix.translate(left_edge - item.matrix[4], 0)


def align_right(elements: set[ElementPresentation]):
    right_edge = _right_edge(elements)

    for item in elements:
        item.matrix.translate(right_edge - (item.matrix[4] + item.width), 0)


def align_vertical_center(elements: set[ElementPresentation]):
    left_edge = _left_edge(elements)
    right_edge = _right_edge(elements)
    center_edge = left_edge + (right_edge - left_edge) / 2

    for item in elements:
        item.matrix.translate(center_edge - item.matrix[4] - item.width / 2, 0)


def align_top(elements: set[ElementPresentation]):
    top_edge = _top_edge(elements)

    for item in elements:
        item.matrix.translate(0, top_edge - item.matrix[5])


def align_bottom(elements: set[ElementPresentation]):
    bottom_edge = _bottom_edge(elements)

    for item in elements:
        item.matrix.translate(0, bottom_edge - (item.matrix[5] + item.height))


def align_horizontal_center(elements: set[ElementPresentation]):
    top_edge = _top_edge(elements)
    bottom_edge = _bottom_edge(elements)
    center_edge = top_edge + (bottom_edge - top_edge) / 2

    for item in elements:
        item.matrix.translate(0, center_edge - item.matrix[5] - item.height / 2)


def resize_max_width(elements: set[ElementPresentation]):
    max_width = _max_width(elements)

    for item in elements:
        item.width = max_width


def resize_max_height(elements: set[ElementPresentation]):
    max_height = _max_height(elements)

    for item in elements:
        item.height = max_height


def resize_max_size(elements: set[ElementPresentation]):
    max_width = _max_width(elements)
    max_height = _max_height(elements)

    for item in elements:
        item.width = max_width
        item.height = max_height


def _left_edge(elements: set[ElementPresentation]):
    return min(item.matrix[4] for item in elements)


def _right_edge(elements: set[ElementPresentation]):
    return max(item.matrix[4] + item.width for item in elements)


def _top_edge(elements: set[ElementPresentation]):
    return min(item.matrix[5] for item in elements)


def _bottom_edge(elements: set[ElementPresentation]):
    return max(item.matrix[5] + item.height for item in elements)


def _max_width(elements: set[ElementPresentation]):
    return max(item.width for item in elements)


def _max_height(elements: set[ElementPresentation]):
    return max(item.height for item in elements)



from __future__ import annotations

import importlib.resources
import textwrap

from gaphor.core.modeling.base import Base, Id, RepositoryProtocol
from gaphor.core.modeling.event import AttributeUpdated, StyleSheetUpdated
from gaphor.core.modeling.properties import attribute
from gaphor.core.styling import CompiledStyleSheet

SYSTEM_STYLE_SHEET = (importlib.resources.files("gaphor") / "diagram.css").read_text(
    "utf-8"
)

DEFAULT_STYLE_SHEET = textwrap.dedent(
    """\
    diagram {
     /* line-style: sloppy 0.3; */
    }
    """
)


class StyleSheet(Base):
    _compiled_style_sheet: CompiledStyleSheet

    def __init__(
        self, id: Id | None = None, model: RepositoryProtocol | None = None
    ) -> None:
        super().__init__(id, model)
        self._instant_style_declarations = ""
        self._system_font_family = "sans"
        self.compile_style_sheet()

    styleSheet: attribute[str] = attribute("styleSheet", str, DEFAULT_STYLE_SHEET)
    naturalLanguage: attribute[str] = attribute("naturalLanguage", str)

    @property
    def instant_style_declarations(self) -> str:
        return self._instant_style_declarations

    @instant_style_declarations.setter
    def instant_style_declarations(self, declarations: str) -> None:
        self._instant_style_declarations = declarations
        self.compile_style_sheet()
        super().handle(StyleSheetUpdated(self))

    @property
    def system_font_family(self) -> str:
        return self._system_font_family

    @system_font_family.setter
    def system_font_family(self, font_family: str) -> None:
        self._system_font_family = font_family
        self.compile_style_sheet()

    def compile_style_sheet(self) -> None:
        self._compiled_style_sheet = CompiledStyleSheet(
            SYSTEM_STYLE_SHEET,
            f"diagram {{ font-family: {self._system_font_family} }}",
            self.styleSheet,
            self._instant_style_declarations,
        )

    def new_compiled_style_sheet(self) -> CompiledStyleSheet:
        return self._compiled_style_sheet.copy()

    def postload(self) -> None:
        super().postload()
        self.compile_style_sheet()

    def handle(self, event: object) -> None:
        # Ensure compiled style sheet is always up-to-date:
        if is_style_sheet_update := (
            isinstance(event, AttributeUpdated)
            and event.property is StyleSheet.styleSheet
        ):
            self.compile_style_sheet()

        super().handle(event)

        if is_style_sheet_update or (
            isinstance(event, AttributeUpdated)
            and event.property is StyleSheet.naturalLanguage
        ):
            super().handle(StyleSheetUpdated(self))



from gaphor.core.modeling.properties import attribute
from gaphor.diagram.presentation import (
    Classified,
    ElementPresentation,
)
from gaphor.diagram.shapes import (
    Box,
    CssNode,
    Text,
    draw_border,
    draw_top_separator,
)
from gaphor.diagram.support import represents
from gaphor.SysML.sysml import InterfaceBlock, ValueType
from gaphor.UML.classes.klass import attributes_compartment, operation_watches
from gaphor.UML.classes.stereotype import stereotype_compartments, stereotype_watches
from gaphor.UML.compartments import name_compartment
from gaphor.UML.umlfmt import format_operation, format_property


@represents(InterfaceBlock)
class InterfaceBlockItem(Classified, ElementPresentation[InterfaceBlock]):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)

        self.watch("show_stereotypes", self.update_shapes).watch(
            "show_parts", self.update_shapes
        ).watch("show_references", self.update_shapes).watch(
            "show_values", self.update_shapes
        ).watch("show_operations", self.update_shapes).watch(
            "subject[UML:NamedElement].name"
        ).watch("subject[UML:NamedElement].name").watch(
            "subject[UML:NamedElement].namespace.name"
        ).watch("subject[Classifier].isAbstract", self.update_shapes).watch(
            "subject[Class].ownedAttribute.aggregation", self.update_shapes
        )
        operation_watches(self, "Block")
        stereotype_watches(self)

    show_stereotypes: attribute[int] = attribute("show_stereotypes", int)

    show_parts: attribute[int] = attribute("show_parts", int, default=False)

    show_references: attribute[int] = attribute("show_references", int, default=False)

    show_values: attribute[int] = attribute("show_values", int, default=False)

    show_attributes: attribute[int] = attribute("show_attributes", int, default=False)

    show_operations: attribute[int] = attribute("show_operations", int, default=False)

    def additional_stereotypes(self):
        from gaphor.RAAML import raaml

        if isinstance(self.subject, raaml.Situation):
            return [self.diagram.gettext("Situation")]
        elif isinstance(self.subject, raaml.ControlStructure):
            return [self.diagram.gettext("ControlStructure")]
        elif isinstance(self.subject, InterfaceBlock):
            return [self.diagram.gettext("interfaceblock")]
        return ()

    def update_shapes(self, event=None):
        self.shape = Box(
            name_compartment(self, self.additional_stereotypes),
            *(
                self.show_references
                and self.subject
                and [
                    self.block_compartment(
                        self.diagram.gettext("references"),
                        lambda a: not a.association and a.aggregation != "composite",
                        "reference",
                    )
                ]
                or []
            ),
            *(
                self.show_values
                and self.subject
                and [
                    self.block_compartment(
                        self.diagram.gettext("values"),
                        lambda a: isinstance(a.type, ValueType)
                        and a.aggregation == "composite",
                        "value",
                    )
                ]
                or []
            ),
            *(
                self.show_attributes
                and self.subject
                and [attributes_compartment(self.subject)]
                or []
            ),
            *(
                self.show_operations
                and self.subject
                and [
                    self.operations_compartment(
                        self.diagram.gettext("operations"), self.subject
                    ),
                ]
                or []
            ),
            *(self.show_stereotypes and stereotype_compartments(self.subject) or []),
            draw=draw_border,
        )

    def block_compartment(self, name, predicate, css_name):
        # We need to fix the attribute value, since the for loop changes it.
        def lazy_format(attribute):
            return lambda: format_property(attribute) or self.diagram.gettext("unnamed")

        return CssNode(
            "compartment",
            self.subject,
            Box(
                CssNode(
                    "heading",
                    self.subject,
                    Text(text=name),
                ),
                *(
                    CssNode(css_name, attribute, Text(text=lazy_format(attribute)))
                    for attribute in self.subject.ownedAttribute
                    if predicate(attribute)
                ),
                draw=draw_top_separator,
            ),
        )

    def operations_compartment(self, name, predicate):
        def lazy_format(operation):
            return lambda: format_operation(operation) or self.diagram.gettext(
                "unnamed"
            )

        return CssNode(
            "compartment",
            self.subject,
            Box(
                CssNode(
                    "heading",
                    self.subject,
                    Text(text=name),
                ),
                *(
                    CssNode("operation", operation, Text(text=lazy_format(operation)))
                    for operation in self.subject.ownedOperation
                ),
                draw=draw_top_separator,
            ),
        )



from __future__ import annotations

import functools
import os
from pathlib import Path
from typing import ClassVar

import sphinx.util.docutils
from docutils import nodes
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives import images
from sphinx.util import logging
from sphinx.util.typing import OptionSpec

import gaphor.storage as storage
from gaphor.core.modeling import Diagram, ElementFactory
from gaphor.diagram.export import save_pdf, save_svg
from gaphor.i18n import gettext
from gaphor.services.modelinglanguage import ModelingLanguageService

log = logging.getLogger(__name__)


def setup(app: sphinx.application.Sphinx) -> dict[str, object]:
    """Called by Sphinx to set up the extension."""
    app.add_config_value("gaphor_models", {}, "env", [dict])

    app.add_directive("diagram", DiagramDirective)

    app.connect("config-inited", config_inited)

    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }


def config_inited(app, config):
    log.info(f"Gaphor models: {config.gaphor_models}")
    if isinstance(config.gaphor_models, str):
        config.gaphor_models = {"default": config.gaphor_models}


class DiagramDirective(sphinx.util.docutils.SphinxDirective):
    """The Gaphor diagram directive.

    Usage: "``.. diagram:: diagram-name``".
    """

    has_content = True
    required_arguments = 1
    final_argument_whitespace = True
    option_spec: ClassVar[OptionSpec] = {
        "model": directives.unchanged,
        **images.Figure.option_spec,
    }

    def run(self) -> list[nodes.Node]:
        name = self.arguments[0]
        model_name = self.options.pop("model", "default")
        model_file = self.config.gaphor_models.get(model_name)

        if not model_file:
            return self.logging_error_node(
                gettext("No model file configured for model '{model_name}'.").format(
                    model_name=model_name
                )
            )

        self.env.note_dependency(model_file)
        model = load_model(Path(self.env.srcdir) / model_file)

        outdir = (Path(self.env.app.doctreedir) / ".." / "gaphor").resolve()
        outdir.mkdir(exist_ok=True)

        diagram = next(
            model.select(
                lambda e: isinstance(e, Diagram)
                and hasattr(e, "qualifiedName")
                and ".".join(e.qualifiedName) == name
            ),
            None,
        )

        if not diagram:
            diagram = next(
                model.select(lambda e: isinstance(e, Diagram) and e.name == name), None
            )

        if not diagram:
            return self.logging_error_node(
                gettext(
                    "No diagram '{name}' in model '{model_name}' ({model_file})."
                ).format(name=name, model_name=model_name, model_file=model_file)
            )

        outfile = outdir / f"{diagram.id}"
        save_svg(outfile.with_suffix(".svg"), diagram)
        save_pdf(outfile.with_suffix(".pdf"), diagram)

        # Image needs a relative path. Make our outfile path relative to the doc
        # os.path.relpath is used because Path.relative_to doesn't support going up
        # to a common parent directory before Python 3.12 with the walk_up option
        outdir = Path(os.path.relpath(outdir, self.env.srcdir))
        outfile = outdir / f"{diagram.id}"

        for _ in Path(self.env.docname).parts[:-1]:
            outfile = Path("..") / outfile

        return images.Figure(  # type: ignore[no-any-return]
            name="diagram/figure",
            arguments=[f"{outfile}.*"],
            options=self.options,
            content=self.content,
            lineno=self.lineno,
            content_offset=self.content_offset,
            block_text=self.block_text,
            state=self.state,
            state_machine=self.state_machine,
        ).run()

    def logging_error_node(self, text: str) -> list[nodes.Node]:
        location = self.state_machine.get_source_and_line(self.lineno)
        log.error(text, location=location)
        return [nodes.error("", nodes.paragraph(text=text))]


@functools.cache
def load_model(model_file: str) -> ElementFactory:
    element_factory = ElementFactory()

    modeling_language = ModelingLanguageService()

    with open(model_file, encoding="utf-8") as file_obj:
        storage.load(
            file_obj,
            element_factory,
            modeling_language,
        )
    return element_factory



import contextlib
import functools

from gaphor import UML
from gaphor.core.format import format, parse
from gaphor.diagram.propertypages import (
    PropertyPageBase,
    PropertyPages,
    help_link,
    unsubscribe_all_on_destroy,
)
from gaphor.transaction import Transaction
from gaphor.UML.classes.association import AssociationItem
from gaphor.UML.classes.classespropertypages import new_builder
from gaphor.UML.profiles.stereotypepropertypages import (
    stereotype_key_handler,
    stereotype_model,
    stereotype_set_model_with_interaction,
)


def blockable(func):
    blocked = 0

    def _blockable(*args, **kwargs):
        if not blocked:
            return func(*args, **kwargs)

    @contextlib.contextmanager
    def block():
        nonlocal blocked
        blocked += 1
        yield
        blocked -= 1

    wrapped = functools.update_wrapper(_blockable, func)
    wrapped.block = block  # type: ignore[attr-defined]
    return wrapped


@PropertyPages.register(UML.Association)
class AssociationPropertyPage(PropertyPageBase):
    NAVIGABILITY = (None, False, True)
    AGGREGATION = UML.Property.aggregation.values

    order = 20

    def __init__(self, subject: UML.Association, event_manager):
        self.subject = subject
        self.event_manager = event_manager
        self.watcher = subject and subject.watcher()
        self.end_name_change_semaphore = 0

    def construct_end(self, builder, end_name, subject):
        title = builder.get_object(f"{end_name}-title")
        if subject.type:
            title.set_text(f"Member End (: {subject.type.name})")

        self.update_end_name(builder, end_name, subject)

        navigation = builder.get_object(f"{end_name}-navigation")
        with self._on_end_navigability_change.block():
            navigation.set_selected(self.NAVIGABILITY.index(subject.navigability))

        aggregation = builder.get_object(f"{end_name}-aggregation")
        with self._on_end_aggregation_change.block():
            aggregation.set_selected(self.AGGREGATION.index(subject.aggregation))

        if stereotypes_model := stereotype_model(subject, self.event_manager):
            stereotype_list = builder.get_object(f"{end_name}-stereotype-list")
            stereotype_set_model_with_interaction(stereotype_list, stereotypes_model)
        else:
            stereotype_frame = builder.get_object(f"{end_name}-stereotype-frame")
            stereotype_frame.set_visible(False)

    @blockable
    def update_end_name(self, builder, end_name, subject):
        name = builder.get_object(f"{end_name}-name")
        new_name = (
            format(
                subject,
                visibility=True,
                is_derived=True,
                multiplicity=True,
            )
            or ""
        )
        with self._on_end_name_change.block():
            name.set_text(new_name)
        return name

    def construct(self):
        if not self.subject or isinstance(self.subject, UML.Extension):
            return

        head_subject = self.subject.memberEnd[0]
        tail_subject = self.subject.memberEnd[1]

        builder = new_builder(
            "association-editor",
            "association-info",
            signals={
                "head-name-changed": (self._on_end_name_change, head_subject),
                "head-navigation-changed": (
                    self._on_end_navigability_change,
                    head_subject,
                ),
                "head-aggregation-changed": (
                    self._on_end_aggregation_change,
                    head_subject,
                ),
                "head-info-clicked": (self._on_association_info_clicked,),
                "head-stereotype-key-pressed": (stereotype_key_handler,),
                "tail-name-changed": (self._on_end_name_change, tail_subject),
                "tail-navigation-changed": (
                    self._on_end_navigability_change,
                    tail_subject,
                ),
                "tail-aggregation-changed": (
                    self._on_end_aggregation_change,
                    tail_subject,
                ),
                "tail-info-clicked": (self._on_association_info_clicked,),
                "tail-stereotype-key-pressed": (stereotype_key_handler,),
            },
        )

        self.info = builder.get_object("association-info")
        help_link(builder, "head-info-icon", "head-info")
        help_link(builder, "tail-info-icon", "tail-info")

        self.construct_end(builder, "head", head_subject)
        self.construct_end(builder, "tail", tail_subject)

        def name_handler(event):
            end_name = "head" if event.element is head_subject else "tail"
            self.update_end_name(builder, end_name, event.element)

        def restore_nav_handler(event):
            for end_name, end_subject in (
                ("head", head_subject),
                ("tail", tail_subject),
            ):
                combo = builder.get_object(f"{end_name}-navigation")
                self._on_end_navigability_change(combo, None, end_subject)

        # Watch on association end:
        if self.watcher:
            self.watcher.watch("memberEnd[Property].name", name_handler).watch(
                "memberEnd[Property].visibility", name_handler
            ).watch("memberEnd[Property].lowerValue", name_handler).watch(
                "memberEnd[Property].upperValue", name_handler
            ).watch(
                "memberEnd[Property].type",
                restore_nav_handler,
            )

        return unsubscribe_all_on_destroy(
            builder.get_object("association-editor"), self.watcher
        )

    @blockable
    def _on_end_name_change(self, entry, subject):
        with self.update_end_name.block():
            with Transaction(self.event_manager, context="editing"):
                parse(subject, entry.get_text())

    @blockable
    def _on_end_navigability_change(self, dropdown, _pspec, subject):
        if subject and subject.opposite and subject.opposite.type:
            with Transaction(self.event_manager):
                UML.recipes.set_navigability(
                    subject.association,
                    subject,
                    self.NAVIGABILITY[dropdown.get_selected()],
                )

    @blockable
    def _on_end_aggregation_change(self, dropdown, _pspec, subject: UML.Property):
        aggregation = self.AGGREGATION[dropdown.get_selected()]
        if aggregation != subject.aggregation:
            with Transaction(self.event_manager):
                subject.aggregation = aggregation

    def _on_association_info_clicked(self, widget, event):
        self.info.set_relative_to(widget)
        self.info.set_visible(True)


@PropertyPages.register(AssociationItem)
class AssociationDirectionPropertyPage(PropertyPageBase):
    order = 20

    def __init__(self, item: AssociationItem, event_manager):
        self.item = item
        self.event_manager = event_manager

    def construct(self):
        if not self.item.subject:
            return None

        builder = new_builder(
            "association-direction-editor",
            signals={
                "show-direction-changed": (self._on_show_direction_change,),
                "invert-direction-changed": (self.on_invert_direction_change,),
            },
        )

        show_direction = builder.get_object("show-direction")
        show_direction.set_active(self.item.show_direction)

        return builder.get_object("association-direction-editor")

    def _on_show_direction_change(self, button, gparam):
        with Transaction(self.event_manager):
            self.item.show_direction = button.get_active()

    def on_invert_direction_change(self, button):
        with Transaction(self.event_manager):
            self.item.invert_direction()



from gi.repository import Gdk, Gtk

from gaphor.diagram.instanteditors import instant_editor, show_popover
from gaphor.transaction import Transaction
from gaphor.UML.general.comment import CommentItem


@instant_editor.register(CommentItem)
def comment_item_editor(item, view, event_manager, pos=None) -> bool:
    def update_text():
        text = buffer.get_text(
            buffer.get_start_iter(), buffer.get_end_iter(), include_hidden_chars=True
        )
        with Transaction(event_manager, context="editing"):
            item.subject.body = text

    subject = item.subject
    if not subject:
        return False

    body = subject.body or ""

    buffer = Gtk.TextBuffer()
    buffer.set_text(body, -1)
    startiter, enditer = buffer.get_bounds()
    buffer.move_mark_by_name("selection_bound", startiter)
    buffer.move_mark_by_name("insert", enditer)

    text_view = Gtk.TextView.new_with_buffer(buffer)
    text_view.set_size_request(100, 50)
    box = view.get_item_bounding_box(item)

    frame = Gtk.Frame()
    frame.set_child(text_view)

    popover = show_popover(frame, view, box, update_text)

    def on_enter(text_view, keyval, keycode, state):
        if keyval in (Gdk.KEY_Return, Gdk.KEY_KP_Enter) and not state & (
            Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK
        ):
            popover.popdown()
            return True

    controller = Gtk.EventControllerKey.new()
    text_view.add_controller(controller)
    controller.connect("key-pressed", on_enter)

    return True



from gaphor.core.modeling import Diagram
from gaphor.diagram.drop import (
    drop,
    drop_on_presentation,
    drop_relationship,
    grow_parent,
)
from gaphor.diagram.group import change_owner, ungroup
from gaphor.diagram.presentation import connect
from gaphor.diagram.support import get_diagram_item, get_diagram_item_metadata
from gaphor.UML.actions.action import (
    AcceptEventActionItem,
    ActionItem,
    CallBehaviorActionItem,
    SendSignalActionItem,
    ValueSpecificationActionItem,
)
from gaphor.UML.actions.objectnode import ObjectNodeItem
from gaphor.UML.actions.partition import PartitionItem
from gaphor.UML.drop import drop_relationship_on_diagram
from gaphor.UML.uml import ActivityEdge, ActivityParameterNode, ActivityPartition


@drop.register(ActivityEdge, Diagram)
def drop_activity_edge_on_diagram(element: ActivityEdge, diagram: Diagram, x, y):
    item_class = get_diagram_item(type(element))
    if not item_class:
        return None

    metadata = get_diagram_item_metadata(item_class)
    return (
        drop_relationship(
            element,
            metadata["head"].get(element),
            metadata["tail"].get(element),
            diagram,
            x,
            y,
        )
        if metadata
        else None
    )


@drop.register(ActivityParameterNode, Diagram)
def drop_parameter_node(element: ActivityParameterNode, diagram: Diagram, x, y):
    item_class = get_diagram_item(type(element))
    if not item_class:
        return None

    def drop_distance_to_item(item):
        local_x, local_y = item.matrix_i2c.inverse().transform_point(x, y)
        return item.point(local_x, local_y)

    appendable_item = min(
        diagram.select(
            lambda item: bool(element.activity and element.activity is item.subject)
        ),
        key=drop_distance_to_item,
        default=None,
    )

    if not appendable_item:
        return None

    item = diagram.create(item_class, subject=element)
    item.matrix.translate(x, y)
    item.change_parent(appendable_item)
    connect(item, item.handles()[0], appendable_item)

    return item


@drop.register(ActivityPartition, Diagram)
def drop_partition(partition: ActivityPartition, diagram: Diagram, x: float, y: float):
    item_class = get_diagram_item(type(partition))
    if not item_class:
        return None

    item = diagram.create(item_class)
    assert item

    item.matrix.translate(x, y)
    item.subject = partition
    item.partition = partition

    for node in partition.node:
        child_class = get_diagram_item(type(node))
        if not child_class:
            continue
        child_item = diagram.create(child_class)
        assert child_item
        child_item.subject = node
        child_item.diagram = diagram
        child_item.matrix.translate(x, y)
        drop_on_presentation(child_item, item, x, y)

    for node in partition.node:
        for outgoing in node.outgoing:
            if not outgoing.target:
                continue
            if partition.node.includes(outgoing.target):
                drop_relationship_on_diagram(outgoing, diagram, x, y)

    return item


@drop.register(ActionItem, PartitionItem)
@drop.register(AcceptEventActionItem, PartitionItem)
@drop.register(SendSignalActionItem, PartitionItem)
@drop.register(CallBehaviorActionItem, PartitionItem)
@drop.register(ValueSpecificationActionItem, PartitionItem)
@drop.register(ObjectNodeItem, PartitionItem)
def drop_on_partition(item, new_parent, x, y):
    assert item.diagram is new_parent.diagram

    if not item.subject:
        return

    old_parent = item.parent
    target_subject = new_parent.partition_at_point((x, y))

    if target_subject is item.subject.inPartition:
        return

    if isinstance(old_parent, PartitionItem) and any(
        ungroup(p, item.subject) for p in old_parent.partition
    ):
        item.change_parent(None)
        old_parent.request_update()

    if change_owner(target_subject, item.subject):
        grow_parent(new_parent, item)
        item.change_parent(new_parent)



"""Load and save Gaphor models to Gaphors own XML format.

Three functions are exported: `load(file_obj)`loads a model from a
file. `save(file_obj)` stores the current model in a file.
"""

__all__ = [
    "load",
    "load_generator",
    "save",
    "save_generator",
    "UnknownModelElementError",
]

from gaphor.storage.load import UnknownModelElementError, load, load_generator
from gaphor.storage.save import save, save_generator



"""UI related events."""


class ElementFocused:
    def __init__(self, element):
        self.element = element


class ElementOpened:
    def __init__(self, element):
        self.element = element


class CurrentDiagramChanged:
    def __init__(self, diagram):
        self.diagram = diagram


class ModelSelectionChanged:
    def __init__(self, service, focused_element):
        self.service = service
        self.focused_element = focused_element


class ToolSelected:
    def __init__(self, tool_name):
        self.tool_name = tool_name



"""The UML Modeling Language module is the entrypoint for UML related
assets."""

from collections.abc import Iterable

from gaphor.abc import ModelingLanguage
from gaphor.core import gettext
from gaphor.diagram.diagramtoolbox import (
    DiagramType,
    ElementCreateInfo,
    ToolboxDefinition,
)
from gaphor.UML import diagramitems, uml
from gaphor.UML.toolbox import uml_diagram_types, uml_element_types, uml_toolbox_actions
from gaphor.UML.treemodel import TreeModel


class UMLModelingLanguage(ModelingLanguage):
    @property
    def name(self) -> str:
        return gettext("UML")

    @property
    def toolbox_definition(self) -> ToolboxDefinition:
        return uml_toolbox_actions

    @property
    def diagram_types(self) -> Iterable[DiagramType]:
        """Return an iterator (id, name) for each diagram type."""
        yield from uml_diagram_types

    @property
    def element_types(self) -> Iterable[ElementCreateInfo]:
        yield from uml_element_types

    @property
    def model_browser_model(self) -> type[TreeModel]:
        return TreeModel

    def lookup_element(self, name, ns=None):
        assert ns in ("UML", None)
        return getattr(uml, name, None) or getattr(diagramitems, name, None)



import logging

from gaphas.model import Model
from gaphas.view import GtkView
from gi.repository import GObject

from gaphor.diagram.selection import Selection
from gaphor.ui.actiongroup import named_shortcut

log = logging.getLogger(__name__)


class DiagramView(GtkView):
    __gtype_name__ = "DiagramView"

    def __init__(self, model: Model | None = None):
        super().__init__(model, Selection())
        self.selection.add_handler(self._on_selection_changed)
        self._on_selection_changed()

    @GObject.Signal(name="cut-clipboard", flags=GObject.SignalFlags.RUN_LAST)
    def _cut_clipboard(self):
        pass

    @GObject.Signal(name="copy-clipboard", flags=GObject.SignalFlags.RUN_LAST)
    def _copy_clipboard(self):
        pass

    @GObject.Signal(name="paste-clipboard", flags=GObject.SignalFlags.RUN_LAST)
    def _paste_clipboard(self):
        pass

    @GObject.Signal(name="paste-full-clipboard", flags=GObject.SignalFlags.RUN_LAST)
    def _paste_full_clipboard(self):
        pass

    @GObject.Signal(name="delete", flags=GObject.SignalFlags.RUN_LAST)
    def _delete(self):
        pass

    @GObject.Signal(name="select-all", flags=GObject.SignalFlags.RUN_LAST)
    def _select_all(self):
        self.selection.select_items(*self.model.get_all_items())

    @GObject.Signal(name="unselect-all", flags=GObject.SignalFlags.RUN_LAST)
    def _unselect_all(self):
        self.selection.unselect_all()

    def _on_selection_changed(self, _item=None):
        enabled = bool(self.selection.selected_items)
        self.action_set_enabled("clipboard.cut", enabled)
        self.action_set_enabled("clipboard.copy", enabled)
        self.action_set_enabled("selection.delete", enabled)
        self.action_set_enabled("selection.unselect-all", enabled)


def _trigger_signal(signal_name):
    def _trigger_action(self, _action_name, _param):
        self.emit(signal_name)

    return _trigger_action


if hasattr(DiagramView, "install_action"):
    DiagramView.set_css_name("diagramview")
    # Deal with Gtk being mocked when generating docs
    DiagramView.install_action("clipboard.cut", None, _trigger_signal("cut-clipboard"))
    DiagramView.install_action(
        "clipboard.copy", None, _trigger_signal("copy-clipboard")
    )
    DiagramView.install_action(
        "clipboard.paste", None, _trigger_signal("paste-clipboard")
    )
    DiagramView.install_action(
        "clipboard.paste-full", None, _trigger_signal("paste-full-clipboard")
    )
    DiagramView.install_action("selection.delete", None, _trigger_signal("delete"))
    DiagramView.install_action(
        "selection.select-all", None, _trigger_signal("select-all")
    )
    DiagramView.install_action(
        "selection.unselect-all", None, _trigger_signal("unselect-all")
    )

    DiagramView.add_shortcut(named_shortcut("<Primary>x", "clipboard.cut"))
    DiagramView.add_shortcut(named_shortcut("<Primary>c", "clipboard.copy"))
    DiagramView.add_shortcut(named_shortcut("<Primary>v", "clipboard.paste"))
    DiagramView.add_shortcut(
        named_shortcut("<Primary><Shift>v", "clipboard.paste-full")
    )
    DiagramView.add_shortcut(
        named_shortcut("Delete|BackSpace|<Meta>BackSpace", "selection.delete")
    )
    DiagramView.add_shortcut(named_shortcut("<Primary>a", "selection.select-all"))
    DiagramView.add_shortcut(
        named_shortcut("<Primary><Shift>a", "selection.unselect-all")
    )



from gaphor.diagram.presentation import LinePresentation, Named, text_name
from gaphor.diagram.shapes import Box, draw_arrow_head
from gaphor.SysML import sysml
from gaphor.UML.compartments import text_stereotypes


class DirectedRelationshipPropertyPathItem(
    Named, LinePresentation[sysml.DirectedRelationshipPropertyPath]
):
    relation_type = ""

    def __init__(self, diagram, id=None):
        super().__init__(
            diagram,
            id,
            shape_middle=Box(
                text_stereotypes(self, lambda: [self.relation_type]),
                text_name(self),
            ),
        )

        self.draw_head = draw_arrow_head
        self.watch("subject[UML:NamedElement].name").watch(
            "subject[UML:Element].appliedStereotype.classifier.name"
        )



from gaphas.tool import item_tool as _item_tool
from gi.repository import Gtk

from gaphor.diagram.event import DiagramSelectionChanged


def item_tool(event_manager) -> Gtk.GestureDrag:
    gesture = _item_tool()
    gesture.connect_after("drag-begin", on_drag_begin, event_manager)
    gesture.connect_after("drag-update", on_drag_update)
    gesture.connect_after("drag-end", on_drag_end)
    return gesture


def on_drag_begin(gesture, _start_x, _start_y, event_manager):
    view = gesture.get_widget()
    selection = view.selection
    event_manager.handle(
        DiagramSelectionChanged(view, selection.focused_item, selection.selected_items)
    )


def on_drag_update(gesture, _offset_x, _offset_y):
    view = gesture.get_widget()
    view.model.update()


def on_drag_end(gesture, _offset_x, _offset_y):
    view = gesture.get_widget()
    view.selection.dropzone_item = None
    view.model.update()



"""Application settings support for Gaphor."""

import hashlib
import logging
import sys
from enum import Enum
from pathlib import Path

from gi.repository import Gio, GLib

APPLICATION_ID = "org.gaphor.Gaphor"

logger = logging.getLogger(__name__)


def get_config_dir() -> Path:
    """Return the directory where the user's config is stored.

    This varies depending on platform.
    """

    config_dir = Path(GLib.get_user_config_dir()) / "gaphor"
    config_dir.mkdir(exist_ok=True, parents=True)

    return config_dir


def get_cache_dir() -> Path:
    """Return the directory where the user's cache is stored.

    This varies depending on platform.
    """

    cache_dir = Path(GLib.get_user_cache_dir()) / "gaphor"
    cache_dir.mkdir(exist_ok=True, parents=True)

    return cache_dir


def file_hash(filename) -> str:
    return hashlib.blake2b(str(filename).encode("utf-8"), digest_size=24).hexdigest()


class StyleVariant(Enum):
    SYSTEM = 0
    DARK = 1
    LIGHT = 2


class Settings:
    """Gaphor settings."""

    _required_keys = [
        "use-english",
        "style-variant",
        "reset-tool-after-create",
        "remove-unused-elements",
    ]

    def __init__(self):
        self._gio_settings = (
            Gio.Settings.new(APPLICATION_ID)
            if (schema_source := Gio.SettingsSchemaSource.get_default())
            and (schema := schema_source.lookup(APPLICATION_ID, False))
            and (schema_keys := schema.list_keys())
            and all(key in schema_keys for key in self._required_keys)
            else None
        )

        if not self._gio_settings:
            # Workaround: do not show this message if we're installing schemas
            if "install-schemas" not in sys.argv:
                logger.warning(
                    "Settings schema not found and settings won’t be saved. Run `gaphor install-schemas`."
                )

    @property
    def style_variant(self) -> StyleVariant:
        return (
            StyleVariant(self._gio_settings.get_enum("style-variant"))
            if self._gio_settings
            else StyleVariant.SYSTEM
        )

    @style_variant.setter
    def style_variant(self, style_variant: StyleVariant):
        if self._gio_settings:
            self._gio_settings.set_enum("style-variant", style_variant.value)

    def bind_style_variant(self, target, prop):
        # Bind with mapping not supported by PyGObject: https://gitlab.gnome.org/GNOME/pygobject/-/issues/98
        # To bind to a function that can map between guint and a string
        if self._gio_settings:
            style_variant = self._gio_settings.get_enum("style-variant")
            target.set_property(prop, style_variant)

    def style_variant_changed(self, callback):
        if self._gio_settings:

            def _on_changed(_settings, _name):
                callback(self.style_variant)

            self._gio_settings.connect("changed::style-variant", _on_changed)
            callback(self.style_variant)

    @property
    def use_english(self) -> bool:
        return (
            self._gio_settings.get_boolean("use-english")
            if self._gio_settings
            else False
        )

    def bind_use_english(self, target, prop):
        self._bind_propery("use-english", target, prop)

    @property
    def reset_tool_after_create(self):
        return (
            self._gio_settings.get_boolean("reset-tool-after-create")
            if self._gio_settings
            else True
        )

    def bind_reset_tool_after_create(self, target, prop):
        self._bind_propery("reset-tool-after-create", target, prop)

    @property
    def remove_unused_elements(self):
        return (
            self._gio_settings.get_boolean("remove-unused-elements")
            if self._gio_settings
            else True
        )

    def bind_remove_unused_elements(self, target, prop):
        self._bind_propery("remove-unused-elements", target, prop)

    def _bind_propery(self, name, target, prop):
        if self._gio_settings:
            self._gio_settings.bind(name, target, prop, Gio.SettingsBindFlags.DEFAULT)


settings = Settings()



from gaphor.C4Model import c4model
from gaphor.diagram.presentation import ElementPresentation, Named, text_name
from gaphor.diagram.shapes import Box, CssNode, Text, ellipse, stroke
from gaphor.diagram.support import represents


@represents(c4model.Database)
class C4DatabaseItem(Named, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)

        self.watch("subject[UML:NamedElement].name")
        self.watch("subject[Container].technology")
        self.watch("subject[Container].description")
        self.watch("subject[Container].type")
        self.watch("children", self.update_shapes)

    def update_shapes(self, event=None):
        diagram = self.diagram
        self.shape = Box(
            text_name(self),
            CssNode(
                "technology",
                self.subject,
                Text(
                    text=lambda: self.subject.technology
                    and f"[{diagram.gettext(self.subject.type)}: {self.subject.technology}]"
                    or f"[{diagram.gettext(self.subject.type)}]",
                ),
            ),
            *(
                ()
                if self.children
                else (
                    CssNode(
                        "description",
                        self.subject,
                        Text(text=lambda: self.subject.description or ""),
                    ),
                )
            ),
            draw=draw_database,
        )


def draw_database(box, context, bounding_box):
    cr = context.cairo
    d = 0.38
    x, y, width, height = bounding_box

    x1 = width + x
    y1 = height + y
    ellipse(cr, x, y - height * d / 2, width, height * d)
    cr.move_to(x1, y)
    cr.line_to(x1, y1)
    cr.curve_to(x1, y1, width / 2, y1 + height * d, x, y1)
    cr.line_to(x, y)

    stroke(context, fill=True)



# ruff: noqa: F401

from gaphor.diagram.general.metadata import MetadataItem
from gaphor.diagram.general.simpleitem import Box, Ellipse, Line

__all__ = [
    "Box",
    "Ellipse",
    "Line",
    "MetadataItem",
]



"""Gaphas connectors.

This module ties connectors in Gaphas -- on presentation -- to model
level connectors (gaphor.connectors).
"""

import logging

from gaphas.connector import Connector as ConnectorAspect
from gaphas.connector import ItemConnector, LineConnector

from gaphor.core.modeling import Presentation
from gaphor.diagram.connectors import (
    Connector,
    ItemConnected,
    ItemDisconnected,
    ItemReconnected,
)
from gaphor.diagram.presentation import ElementPresentation, LinePresentation

log = logging.getLogger(__name__)


@ConnectorAspect.register(Presentation)
@ConnectorAspect.register(ElementPresentation)
class PresentationConnector(ItemConnector):
    """Handle Tool (acts on item handles) that uses the Connector protocol to
    connect items to one-another.

    It also adds handles to lines when a line is grabbed in the middle
    of a line segment (points are drawn by the LineSegmentPainter).
    """

    def allow(self, sink):
        adapter = Connector(sink.item, self.item)
        return adapter and adapter.allow(self.handle, sink.port)

    def connect(self, sink):
        """Create connection at handle level and at model level."""
        handle = self.handle
        item = self.item
        cinfo = self.connections.get_connection(handle)

        if cinfo and cinfo.connected is sink.item:
            self.reconnect_handle(sink)
            return

        adapter = Connector(sink.item, item)
        if cinfo:
            self.disconnect()

        self.glue(sink)
        if not sink.port:
            log.warning("No port found", item, sink.item)
            return

        self.connect_handle(sink)
        adapter.connect(handle, sink.port)

    def connect_handle(self, sink):
        super().connect_handle(sink, callback=DisconnectHandle())
        item = self.item
        item.handle(ItemConnected(item, self.handle, sink.item, sink.port))

    def disconnect(self):
        # Model level disconnect and event is handled in callback
        super().disconnect()

    def disconnect_handle(self):
        if cinfo := self.connections.get_connection(self.handle):
            # prevent model level disconnect from triggering
            cinfo.callback.disable = True
        super().disconnect()

    def reconnect_handle(self, sink):
        """Reconnect only constraint - leave model intact."""
        handle = self.handle
        item = self.item
        log.debug("performing reconnect constraint")
        self.glue(sink)
        constraint = sink.constraint(item, handle)
        self.connections.reconnect_item(item, handle, sink.port, constraint=constraint)
        item.handle(ItemReconnected(item, handle, sink.item, sink.port))


@ConnectorAspect.register(LinePresentation)
class LinePresentationConnector(PresentationConnector, LineConnector):
    pass


class DisconnectHandle:
    """Callback for items disconnection using the adapters.

    This is an object so disconnection data can be serialized/deserialized
    using pickle.

    Args:
      item (Item): Connecting item.
      handle (Handle): Handle of connecting item.
      connections (Connections): Connections object containing connection constraints.
    """

    def __init__(self):
        self.disable = False

    def __call__(self, item, handle, connected, port):
        if self.disable:
            log.debug(f"Disconnect callback disabled for {item}.{handle} (disabled)")
        else:
            log.debug(f"Disconnect callback {item}.{handle}")
            adapter = Connector(connected, item)
            adapter.disconnect(handle)
        item.handle(ItemDisconnected(item, handle, connected, port))



"""Service dedicated to exporting diagrams to a variety of file formats."""

import contextlib
import re

import cairo
from gaphas.geometry import Rectangle
from gaphas.painter import FreeHandPainter, PainterChain

from gaphor.core.modeling.diagram import StyledDiagram
from gaphor.diagram.painter import DiagramTypePainter, ItemPainter


def escape_filename(diagram_name):
    return re.sub("\\W+", "_", diagram_name)


def render(diagram, new_surface, items=None, with_diagram_type=True, padding=8) -> None:
    if items is None:
        items = list(diagram.get_all_items())

    diagram.update(diagram.ownedPresentation)

    painter = new_painter(diagram)
    if with_diagram_type:
        painter.append(DiagramTypePainter(diagram))
        type_padding = diagram_type_height(diagram)
    else:
        type_padding = 0

    # Update bounding boxes with a temporary Cairo Context
    # (used for stuff like calculating font metrics)
    bounding_box = calc_bounding_box(items, painter)

    w, h = (
        bounding_box.width + 2 * padding,
        bounding_box.height + 2 * padding + type_padding,
    )

    with new_surface(w, h) as surface:
        cr = cairo.Context(surface)

        bg_color = diagram.style(StyledDiagram(diagram)).get("background-color")
        if bg_color and bg_color[3]:
            cr.rectangle(0, 0, w, h)
            cr.set_source_rgba(*bg_color)
            cr.fill()

        cr.translate(
            -bounding_box.x + padding, -bounding_box.y + padding + type_padding
        )
        painter.paint(items, cr)
        cr.show_page()
        surface.flush()


def diagram_type_height(diagram):
    if not diagram.diagramType:
        return 0

    bounding_box = calc_bounding_box(
        diagram.get_all_items(), DiagramTypePainter(diagram)
    )

    return bounding_box.height


def calc_bounding_box(items, painter):
    surface = cairo.RecordingSurface(cairo.Content.COLOR_ALPHA, None)
    cr = cairo.Context(surface)
    painter.paint(items, cr)
    return Rectangle(*surface.ink_extents())


def save_svg(filename, diagram):
    render(diagram, lambda w, h: cairo.SVGSurface(filename, w, h))


def save_png(filename, diagram):
    @contextlib.contextmanager
    def new_png_surface(w, h):
        with cairo.ImageSurface(cairo.FORMAT_ARGB32, int(w + 1), int(h + 1)) as surface:
            yield surface
            surface.write_to_png(filename)

    render(
        diagram,
        new_png_surface,
    )


def save_pdf(filename, diagram):
    render(diagram, lambda w, h: cairo.PDFSurface(filename, w, h))


def save_eps(filename, diagram):
    def new_surface(w, h):
        surface = cairo.PSSurface(filename, w, h)
        surface.set_eps(True)
        return surface

    render(diagram, new_surface)


def new_painter(diagram):
    style = diagram.style(StyledDiagram(diagram))
    sloppiness = style.get("line-style", 0.0)
    item_painter = (
        FreeHandPainter(ItemPainter(), sloppiness) if sloppiness else ItemPainter()
    )
    return PainterChain().append(item_painter)



from collections.abc import Callable

from gaphas.geometry import Rectangle

from gaphor.core.modeling import DrawContext, Presentation
from gaphor.diagram.presentation import text_name
from gaphor.diagram.shapes import Box, CssNode, Shape, Text
from gaphor.UML.recipes import stereotypes_str


def name_compartment(
    presentation: Presentation,
    additional_stereotypes: Callable[[], list[str]] | None = None,
    draw_icon: Callable[[Box, DrawContext, Rectangle], None] | None = None,
):
    return CssNode(
        "compartment",
        None,
        Box(
            text_stereotypes(presentation, additional_stereotypes),
            text_name(presentation),
            text_from_package(presentation),
            draw=draw_icon,
        ),
    )


def text_stereotypes(
    item: Presentation, additional_stereotypes: Callable[[], list[str]] | None = None
) -> Shape:
    return CssNode(
        "stereotypes",
        item.subject,
        Text(
            text=lambda: stereotypes_str(
                item.subject, additional_stereotypes() if additional_stereotypes else ()
            ),
        ),
    )


def text_from_package(item: Presentation):
    return CssNode("from", item.subject, Text(text=lambda: from_package_str(item)))


def from_package_str(item):
    """Display name space info when it is different, then diagram's or parent's
    namespace."""
    subject = item.subject
    diagram = item.diagram

    if not (subject and diagram):
        return False

    namespace = subject.namespace
    parent = item.parent

    if parent and parent.subject and parent.subject is namespace:
        return False

    return (
        diagram.gettext("(from {namespace})").format(namespace=namespace.name)
        if namespace is not item.diagram.owner
        else ""
    )



from gaphas.tool.itemtool import item_at_point
from gi.repository import Gdk, GObject, Gtk

from gaphor.diagram.diagramtoolbox import get_tool_def
from gaphor.diagram.drop import drop
from gaphor.diagram.group import can_group
from gaphor.diagram.tools.placement import create_item
from gaphor.event import Notification
from gaphor.i18n import gettext
from gaphor.transaction import Transaction


class ElementDragData(GObject.Object):
    elements = GObject.Property(type=object)


class ToolboxActionDragData(GObject.Object):
    action = GObject.Property(type=str)


def drop_target_tool(modeling_language, event_manager) -> Gtk.EventController:
    """DropTarget tool."""
    drop_target = Gtk.DropTarget.new(GObject.TYPE_NONE, Gdk.DragAction.COPY)
    drop_target.set_gtypes([ElementDragData.__gtype__, ToolboxActionDragData.__gtype__])
    drop_target.set_preload(True)
    drop_target.connect("motion", on_motion, modeling_language)
    drop_target.connect("drop", on_drop, modeling_language, event_manager)
    return drop_target


def on_motion(
    target: Gtk.DropTarget, x: float, y: float, modeling_language
) -> Gdk.DragAction:
    view = target.get_widget()
    source_value = target.get_value()

    if view.selection.dropzone_item:
        view.model.request_update(view.selection.dropzone_item)

    if not (
        (parent_item := next(item_at_point(view, (x, y)), None)) and parent_item.subject
    ):
        view.selection.dropzone_item = None
        return Gdk.DragAction.COPY

    if (
        isinstance(source_value, ElementDragData)
        and (len(source_value.elements) == 1)
        and can_group(parent_item.subject, source_value.elements[0])
    ):
        view.selection.dropzone_item = parent_item
        view.model.request_update(parent_item)
    elif (
        isinstance(source_value, ToolboxActionDragData)
        and (tool_def := get_tool_def(modeling_language, source_value.action))
        and can_group(parent_item.subject, tool_def.item_factory.subject_class)
    ):
        view.selection.dropzone_item = parent_item
        view.model.request_update(parent_item)
    else:
        view.selection.dropzone_item = None

    return Gdk.DragAction.COPY


def on_drop(target, source_value, x, y, modeling_language, event_manager):
    view = target.get_widget()
    if isinstance(source_value, ElementDragData):
        elements = source_value.elements
        new_parent = view.selection.dropzone_item
        view.selection.dropzone_item = None

        with Transaction(event_manager):
            items = []
            for element in elements:
                if item := _drop(view, element, new_parent, x, y):
                    x += 20
                    y += 20
                    items.append(item)

            if len(items) > 1:
                view.selection.unselect_all()
                view.selection.select_items(*items)
                return True
            if items:
                view.selection.unselect_all()
                view.selection.focused_item = item
                return True
            else:
                event_manager.handle(
                    Notification(gettext("Element can’t be represented on a diagram."))
                )
        return True
    elif isinstance(source_value, ToolboxActionDragData):
        tool_def = get_tool_def(modeling_language, source_value.action)
        with Transaction(event_manager):
            create_item(view, tool_def.item_factory, event_manager, x, y)
        return True

    return False


def _drop(view, item, new_parent, x, y):
    return (
        drop(
            item,
            new_parent,
            *view.get_matrix_v2i(new_parent).transform_point(x, y),
        )
        if new_parent
        else drop(
            item,
            view.model,
            *view.matrix.inverse().transform_point(x, y),
        )
    )



from gaphas.connector import Handle
from gaphas.guide import GuidedItemHandleMoveMixin
from gaphas.handlemove import ConnectionSinkType, HandleMove, ItemHandleMove
from gaphas.item import Item, Line
from gaphas.move import Move
from gaphas.types import Pos
from gaphas.view import GtkView

from gaphor.diagram.connectors import Connector, ItemTemporaryDisconnected
from gaphor.diagram.presentation import AttachedPresentation, connect


def connectable(line, handle, element):
    connector = Connector(element, line)
    return any(connector.allow(handle, port) for port in element.ports())


class GrayOutLineHandleMoveMixin:
    view: GtkView
    item: Item
    handle: Handle

    def start_move(self, pos):
        super().start_move(pos)  # type: ignore[misc]
        handle = self.handle
        if handle.connectable:
            line = self.item
            model = self.view.model
            selection = self.view.selection
            selection.grayed_out_items = {
                item
                for item in model.get_all_items()
                if not (item is line or connectable(line, handle, item))
            }

    def stop_move(self, pos):
        super().stop_move(pos)  # type: ignore[misc]
        selection = self.view.selection
        selection.grayed_out_items = set()
        selection.dropzone_item = None

    def glue(
        self, pos: Pos, distance: int = ItemHandleMove.GLUE_DISTANCE
    ) -> ConnectionSinkType | None:
        sink = super().glue(pos, distance)  # type: ignore[misc]
        self.view.selection.dropzone_item = sink and sink.item
        return sink

    def connect(self, pos: Pos) -> None:
        super().connect(pos)  # type: ignore[misc]
        self.view.selection.dropzone_item = None


class DisconnectEventHandleMoveMixin:
    view: GtkView
    item: Item
    handle: Handle

    def start_move(self, pos: Pos) -> None:
        super().start_move(pos)  # type: ignore[misc]
        model = self.view.model
        assert model
        if cinfo := model.connections.get_connection(self.handle):
            self.item.handle(
                ItemTemporaryDisconnected(
                    cinfo.item, cinfo.handle, cinfo.connected, cinfo.port
                )
            )


@HandleMove.register(Line)
class LineHandleMove(
    GrayOutLineHandleMoveMixin,
    DisconnectEventHandleMoveMixin,
    GuidedItemHandleMoveMixin,
    ItemHandleMove,
):
    pass


@Move.register(AttachedPresentation)
def attached_presentation_move(item, view):
    return LineHandleMove(item, item.handles()[0], view)


class StickyAttachedHandleMove(GuidedItemHandleMoveMixin, ItemHandleMove):
    """Allows item to be connected to its parent by dragging.
    Once a connection is estableshed, the item is stuck with its parent
    and moves only along its border.
    """

    def connect(self, pos: Pos) -> None:
        if self.item.parent:
            self.snap_to_parent()
        else:
            super().connect(pos)

    def move(self, pos: Pos) -> None:
        super().move(pos)

        if self.item.parent:
            self.snap_to_parent()

    def snap_to_parent(self):
        assert self.item.parent
        connect(self.item, self.handle, self.item.parent)
        self.view.model.request_update(self.item)


def sticky_attached_move(item, view):
    return StickyAttachedHandleMove(item, item.handles()[0], view)



"""The main application window."""

from __future__ import annotations

import logging
from collections.abc import Callable
from pathlib import Path

from gi.repository import Gio, GLib, Gtk

from gaphor.abc import ActionProvider, Service
from gaphor.application import distribution
from gaphor.core import action, event_handler, gettext
from gaphor.core.modeling import ModelReady
from gaphor.event import (
    ActionEnabled,
    ActiveSessionChanged,
    ModelSaved,
    Notification,
    SessionCreated,
    SessionShutdownRequested,
    TransactionClosed,
)
from gaphor.i18n import translated_ui_string
from gaphor.services.modelinglanguage import ModelingLanguageChanged
from gaphor.services.undomanager import UndoManagerStateChanged
from gaphor.ui import macos_menubar
from gaphor.ui.abc import UIComponent
from gaphor.ui.actiongroup import window_action_group
from gaphor.ui.event import CurrentDiagramChanged
from gaphor.ui.filedialog import pretty_path
from gaphor.ui.modelbrowser import create_diagram_types_model
from gaphor.ui.notification import InAppNotifier

log = logging.getLogger(__name__)


def new_builder():
    builder = Gtk.Builder()
    builder.add_from_string(translated_ui_string("gaphor.ui", "mainwindow.ui"))
    return builder


def create_modeling_language_model(modeling_language):
    model = Gio.Menu.new()
    for id, name in modeling_language.modeling_languages:
        menu_item = Gio.MenuItem.new(name, "win.select-modeling-language")
        menu_item.set_attribute_value("target", GLib.Variant.new_string(id))
        model.append_item(menu_item)
    return model


class MainWindow(Service, ActionProvider):
    """The main window for the application.

    It contains a Namespace-based tree view and a menu and a statusbar.
    """

    def __init__(
        self,
        event_manager,
        component_registry,
        properties,
        modeling_language,
        export_menu,
        tools_menu,
    ):
        self.event_manager = event_manager
        self.component_registry = component_registry
        self.properties = properties
        self.modeling_language = modeling_language
        self.export_menu = export_menu
        self.tools_menu = tools_menu

        self._builder: Gtk.Builder | None = new_builder()
        self.window: Gtk.Window | None = None
        self.modeling_language_name = None
        self.diagram_types = None
        self.in_app_notifier = None

        # UI updates to be performed when the window is shown/open.
        self._ui_updates: list[Callable[[], None]] = []

        event_manager.subscribe(self._on_file_manager_state_changed)
        event_manager.subscribe(self._on_current_diagram_changed)
        event_manager.subscribe(self._on_model_ready)
        event_manager.subscribe(self._on_undo_manager_state_changed)
        event_manager.subscribe(self._on_action_enabled)
        event_manager.subscribe(self._on_modeling_language_selection_changed)
        event_manager.subscribe(self._on_notification)

    def shutdown(self):
        if self.window:
            self.window.destroy()
            self._builder = None
            self.in_app_notifier = None

        em = self.event_manager
        em.unsubscribe(self._on_file_manager_state_changed)
        em.unsubscribe(self._on_current_diagram_changed)
        em.unsubscribe(self._on_model_ready)
        em.unsubscribe(self._on_undo_manager_state_changed)
        em.unsubscribe(self._on_action_enabled)
        em.unsubscribe(self._on_modeling_language_selection_changed)
        em.unsubscribe(self._on_notification)

    @property
    def title(self):
        return self._builder.get_object("title") if self._builder else None

    @property
    def modified(self):
        return self._builder.get_object("modified") if self._builder else None

    @property
    def subtitle(self):
        return self._builder.get_object("subtitle") if self._builder else None

    @property
    def element_editor_overlay(self):
        return (
            self._builder.get_object("element-editor-overlay")
            if self._builder
            else None
        )

    @property
    def model_changed(self) -> bool:
        return self.modified.get_visible() if self.modified else False

    @model_changed.setter
    def model_changed(self, model_changed: bool):
        if self.modified:
            self.modified.set_visible(model_changed)

    def open(self, gtk_app=None):
        """Open the main window."""

        builder = self._builder
        assert builder

        if not self.window:
            self.window = builder.get_object("main-window")

        window = self.window
        window.set_application(gtk_app)
        if ".dev" in distribution().version:
            window.add_css_class("devel")

        select_modeling_language = builder.get_object("select-modeling-language")
        select_modeling_language.set_menu_model(
            create_modeling_language_model(self.modeling_language)
        )
        self.modeling_language_name = builder.get_object("modeling-language-name")

        self.diagram_types = builder.get_object("diagram-types")
        self.diagram_types.set_menu_model(
            create_diagram_types_model(self.modeling_language)
        )

        if macos_menubar():
            builder.get_object("hamburger-menu-button").unparent()
        else:
            builder.get_object("export-menu").append_submenu(
                gettext("Export"), self.export_menu.menu
            )
            builder.get_object("tools-menu").append_submenu(
                gettext("Tools"), self.tools_menu.menu
            )

        window.set_default_size(*(self.properties.get("ui.window-size", (1024, 640))))
        if self.properties.get("ui.window-mode", "") == "maximized":
            window.maximize()
        elif self.properties.get("ui.window-mode", "") == "fullscreened":
            window.fullscreen()

        track_paned_position(
            builder.get_object("left-pane"), "ui.namespace-width", self.properties
        )

        track_paned_position(
            builder.get_object("top-left-pane"), "ui.namespace-height", self.properties
        )

        if overlay := self.element_editor_overlay:
            overlay.set_show_sidebar(self.properties.get("show-editors", True))

        for name, component in self.component_registry.all(UIComponent):
            if bin := builder.get_object(f"component:{name}"):
                widget = component.open()
                widget.set_name(name)
                bin.set_child(widget)

        action_groups = {
            "win": window_action_group(self.component_registry),
            "text": window_action_group(self.component_registry, scope="text"),
        }
        for scope, action_group in action_groups.items():
            window.insert_action_group(scope, action_group)

        self._on_modeling_language_selection_changed()
        self.in_app_notifier = InAppNotifier(builder.get_object("main-overlay"))

        window.connect("close-request", self._on_window_close_request)
        window.connect("notify::default-height", self._on_window_size_changed)
        window.connect("notify::default-width", self._on_window_size_changed)
        window.connect("notify::maximized", self._on_window_mode_changed)
        window.connect("notify::fullscreened", self._on_window_mode_changed)
        window.connect("notify::is-active", self._on_window_active)
        window.connect("notify::focus-widget", self._on_window_focus_widget)
        window.present()

        for handler in self._ui_updates:
            handler()
        del self._ui_updates[:]

    # Actions:

    @action(
        name="show-editors",
        shortcut="F9",
        state=lambda self: self.properties.get("show-editors", True),
    )
    def toggle_editor_visibility(self, active):
        if overlay := self.element_editor_overlay:
            overlay.set_show_sidebar(active)
        self.properties.set("show-editors", active)

    @action("maximize", state=lambda self: self.window.is_maximized())
    def toggle_maximized(self, active):
        if not self.window:
            return

        if active:
            self.window.maximize()
        else:
            self.window.unmaximize()

    @action(
        "fullscreen", shortcut="F11", state=lambda self: self.window.is_fullscreen()
    )
    def toggle_fullscreen(self, active):
        if not self.window:
            return

        if active:
            self.window.fullscreen()
            self.event_manager.handle(
                Notification(gettext("Press F11 to exit Fullscreen"))
            )
        else:
            self.window.unfullscreen()

    # Signal callbacks:

    @event_handler(SessionCreated, ModelSaved)
    def _on_file_manager_state_changed(
        self, event: SessionCreated | ModelSaved
    ) -> None:
        if not (window := self.window):
            self._ui_updates.append(lambda: self._on_file_manager_state_changed(event))
            return

        filename = Path(event.filename) if event.filename else None

        self.subtitle.set_text(
            pretty_path(filename) if filename else gettext("New model")
        )
        window.set_title(
            f"{filename.name} ({pretty_path(filename.parent)}) - Gaphor"
            if filename
            else f"{gettext('New model')} - Gaphor"
        )

        self.model_changed = False

        window.present()

    @event_handler(ModelReady)
    def _on_model_ready(self, event: ModelReady):
        self.model_changed = event.modified

    @event_handler(CurrentDiagramChanged)
    def _on_current_diagram_changed(self, event):
        self.title.set_text(
            (event.diagram.name or gettext("<None>")) if event.diagram else "Gaphor"
        )

    @event_handler(UndoManagerStateChanged)
    def _on_undo_manager_state_changed(self, event: UndoManagerStateChanged):
        self.model_changed = True

    @event_handler(ActionEnabled)
    def _on_action_enabled(self, event):
        if self.window:
            self.window.action_set_enabled(event.action_name, event.enabled)
            if self._builder and event.action_name == "win.diagram-align":
                self._builder.get_object("alignment-button").set_sensitive(
                    event.enabled
                )
        else:
            self._ui_updates.append(lambda: self._on_action_enabled(event))

    @event_handler(ModelingLanguageChanged)
    def _on_modeling_language_selection_changed(self, event=None):
        if self.modeling_language_name:
            self.modeling_language_name.set_label(self.modeling_language.name)
        if self.diagram_types:
            self.diagram_types.set_menu_model(
                create_diagram_types_model(self.modeling_language)
            )

    @event_handler(Notification)
    def _on_notification(self, event: Notification):
        if not self.in_app_notifier:
            self._ui_updates.append(lambda: self._on_notification(event))
            return

        self.in_app_notifier.handle(event)

    def _on_window_active(self, window, _prop):
        app = window.get_application()
        app.update_menu("export", self.export_menu.menu)
        app.update_menu("tools", self.tools_menu.menu)
        self.event_manager.handle(ActiveSessionChanged(self))

    def _on_window_focus_widget(self, _window, _prop):
        self.event_manager.handle(TransactionClosed())

    def _on_window_close_request(self, _window, _event=None):
        self.event_manager.handle(SessionShutdownRequested())
        return True

    def _on_window_size_changed(self, window, _gspec):
        if not is_maximized(window):
            width, height = window.get_default_size()
            self.properties.set("ui.window-size", (width, height))

    def _on_window_mode_changed(self, window, gspec):
        mode = gspec.name
        self.properties.set("ui.window-mode", mode if window.get_property(mode) else "")


def is_maximized(window: Gtk.Window) -> bool:
    return window.is_maximized() or window.is_fullscreen()  # type: ignore[no-any-return]


def _paned_position_changed(paned, _gparam, name, properties):
    if not is_maximized(paned.get_root()):
        properties.set(name, paned.props.position)


def _paned_ensure_visible(paned, _gparam):
    if paned.props.position < paned.props.min_position + 12:
        paned.props.position = paned.props.min_position + 12
    elif paned.props.position > paned.props.max_position - 12:
        paned.props.position = paned.props.max_position - 12


def track_paned_position(paned, name, properties):
    paned.connect("notify::position", _paned_ensure_visible)
    paned.connect("notify::min-position", _paned_ensure_visible)
    paned.connect("notify::max-position", _paned_ensure_visible)

    if position := properties.get(name, 0):
        paned.set_position(position)

    paned.connect("notify::position", _paned_position_changed, name, properties)



"""Object node item."""

from gaphor import UML
from gaphor.core.modeling.properties import attribute
from gaphor.diagram.presentation import ElementPresentation, Named
from gaphor.diagram.shapes import Box, CssNode, IconBox, Text, draw_border
from gaphor.diagram.support import represents
from gaphor.i18n import i18nize
from gaphor.UML.compartments import text_stereotypes

DEFAULT_UPPER_BOUND = "*"

ORDERING_TEXT = {
    "unordered": i18nize("unordered"),
    "ordered": i18nize("ordered"),
    "LIFO": i18nize("LIFO"),
    "FIFO": i18nize("FIFO"),
}


@represents(UML.ObjectNode)
class ObjectNodeItem(Named, ElementPresentation):
    """Representation of object node. Object node is ordered and has upper
    bound specification.

    Ordering information can be hidden by user.
    """

    def __init__(self, diagram, id=None):
        super().__init__(
            diagram,
            id,
            shape=IconBox(
                Box(
                    text_stereotypes(self),
                    CssNode("name", self.subject, Text(text=self._format_name)),
                    draw=draw_border,
                ),
                CssNode(
                    "upperbound",
                    None,
                    Text(
                        text=lambda: self.subject.upperBound
                        not in (None, "", DEFAULT_UPPER_BOUND)
                        and f"{{ {diagram.gettext('upperBound')} = {self.subject.upperBound} }}"
                        or ""
                    ),
                ),
                CssNode(
                    "ordering",
                    None,
                    Text(
                        text=lambda: self.show_ordering
                        and self.subject.ordering
                        and f"{{ {diagram.gettext('ordering')} = {diagram.gettext(ORDERING_TEXT.get(self.subject.ordering))} }}"
                        or ""
                    ),
                ),
            ),
            width=50,
            height=30,
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject[Element].appliedStereotype.classifier.name")
        self.watch("subject[TypedElement].type.name")
        self.watch("subject[ObjectNode].upperBound")
        self.watch("subject[ObjectNode].ordering")
        self.watch("show_type")
        self.watch("show_ordering")

    show_type: attribute[int] = attribute("show_type", int, default=False)
    show_ordering: attribute[int] = attribute("show_ordering", int, default=False)

    def load(self, name, value):
        if name == "show-ordering":
            name = "show_ordering"
        super().load(name, value)

    def _format_name(self):
        if subject := self.subject:
            if self.show_type and subject.type:
                return f"{subject.name or ''}: {subject.type.name or ''}"
            return subject.name or ""

        return ""



from __future__ import annotations

from collections.abc import Iterator
from typing import NamedTuple

from gaphor.core.modeling import Id
from gaphor.diagram.copypaste import (
    BaseCopy,
    Opaque,
    copy,
    copy_diagram,
    copy_element,
    paste,
    paste_element,
)
from gaphor.UML.recipes import owner_package
from gaphor.UML.uml import (
    Diagram,
    MultiplicityElement,
    NamedElement,
    Package,
    Relationship,
    Slot,
    Type,
)


def copy_multiplicity(
    element: MultiplicityElement, blacklist: list[str] | None = None
) -> Iterator[tuple[Id, Opaque]]:
    if element.lowerValue:
        yield from copy(element.lowerValue)
    if element.upperValue:
        yield from copy(element.upperValue)


class NamedElementCopy(NamedTuple):
    element_copy: BaseCopy


def copy_named_element(
    element: NamedElement, blacklist: list[str] | None = None
) -> NamedElementCopy:
    return NamedElementCopy(
        element_copy=copy_element(element, blacklist),
    )


@copy.register
def _copy_named_element(element: NamedElement) -> Iterator[tuple[Id, NamedElementCopy]]:
    yield element.id, copy_named_element(element)


@copy.register
def _copy_diagram(element: Diagram) -> Iterator[tuple[Id, Opaque]]:
    yield from copy_diagram(element)


def paste_named_element(copy_data: NamedElementCopy, diagram, lookup):
    paster = paste_element(
        copy_data.element_copy,
        diagram,
        lookup,
        filter=lambda n, v: not isinstance(v, Relationship),
    )
    element = next(paster)
    yield element
    next(paster, None)
    if (
        isinstance(element, Type | Package)
        and (not element.namespace)
        and (package := owner_package(diagram.owner)) is not element
    ):
        if isinstance(element, Package):
            element.nestingPackage = package
        else:
            element.package = package


paste.register(NamedElementCopy, paste_named_element)


@copy.register
def copy_slot(element: Slot):
    yield element.id, copy_element(element)
    if element.value:
        yield from copy(element.value)



from gaphor.diagram.copypaste import copy
from gaphor.UML import Constraint, State, Transition
from gaphor.UML.copypaste import copy_named_element


@copy.register
def copy_state(element: State):
    yield element.id, copy_named_element(element)
    if element.entry:
        yield from copy(element.entry)
    if element.exit:
        yield from copy(element.exit)
    if element.doActivity:
        yield from copy(element.doActivity)


@copy.register
def copy_transition(element: Transition):
    yield element.id, copy_named_element(element)
    if element.guard:
        yield from copy(element.guard)


@copy.register
def copy_constraint(element: Constraint):
    yield element.id, copy_named_element(element)
    if element.specification:
        yield from copy(element.specification)



"""Definitions (types) for style sheets."""

from collections.abc import Callable, Sequence
from enum import Enum
from typing import NamedTuple

import tinycss2.color3
from tinycss2.ast import FunctionBlock
from tinycss2.parser import parse_declaration_list

Number = int | float
Color = tuple[float, float, float, float]  # RGBA
Padding = tuple[Number, Number, Number, Number]  # top/right/bottom/left


class TextAlign(Enum):
    LEFT = "left"
    CENTER = "center"
    RIGHT = "right"


class VerticalAlign(Enum):
    TOP = "top"
    MIDDLE = "middle"
    BOTTOM = "bottom"


class JustifyContent(Enum):
    CENTER = "center"
    END = "end"
    START = "start"
    STRETCH = "stretch"


class FontStyle(Enum):
    NORMAL = "normal"
    ITALIC = "italic"


class FontWeight(Enum):
    NORMAL = "normal"
    BOLD = "bold"


class TextDecoration(Enum):
    NONE = "none"
    UNDERLINE = "underline"


class WhiteSpace(Enum):
    NORMAL = "normal"
    NOWRAP = "nowrap"


class Var(NamedTuple):
    name: str


FONT_SIZE_VALUES: dict[str, int | float] = {
    "x-small": 3 / 4,
    "small": 8 / 9,
    "medium": 1,
    "large": 6 / 5,
    "x-large": 3 / 2,
}


def parse_declarations(declaration_list):
    for decl in parse_declaration_list(
        declaration_list, skip_comments=True, skip_whitespace=True
    ):
        if decl.type == "declaration":
            name = decl.lower_name
            yield (
                name,
                declarations(name, parse_value(decl.value)),
            )
        elif decl.type == "error":
            yield "error", decl


class _Declarations:
    """Convert raw CSS declarations into Gaphor styling declarations."""

    def __init__(self) -> None:
        self.declarations: list[tuple[str, Callable[[str, object], object | None]]] = []

    def register(self, *properties):
        def reg(func):
            for prop in properties:
                self.declarations.append((prop, func))
            return func

        return reg

    def __call__(self, property, value):
        if property.startswith("--"):
            return value
        if var := parse_var(value):
            return var

        for prop, func in self.declarations:
            if property == prop:
                return func(property, value)


declarations = _Declarations()


def parse_var(value):
    if isinstance(value, FunctionBlock) and value.name == "var" and value.arguments:
        for arg in value.arguments:
            if arg.type in ("whitespace", "comment"):
                pass
            elif arg.type in ("ident", "string"):
                return Var(arg.value)
    return None


def parse_value(tokens):
    value = []

    for token in tokens:
        if token.type in ("whitespace", "comment"):
            pass
        elif token.type in ("ident", "string"):
            value.append(token.value)
        elif token.type in ("dimension", "number"):
            value.append(token.int_value if token.is_integer else token.value)
        elif token.type == "hash":
            value.append(f"#{token.value}")
        elif token.type == "function":
            value.append(token)
        else:
            value.append(token.serialize())

    return value[0] if len(value) == 1 else tuple(value)


number = (int, float)


def _clip_color(c):
    return 0 if c < 0 else min(c, 1)


@declarations.register("background-color", "color", "text-color")
def parse_color(prop, value):
    try:
        color = tinycss2.color3.parse_color(value)
    except AttributeError:
        return None
    return tuple(_clip_color(v) for v in color) if color else None


@declarations.register(
    "min-width",
    "max-width",
    "min-height",
    "line-width",
    "vertical-spacing",
    "border-radius",
)
def parse_positive_number(prop, value) -> Number | None:
    return value if isinstance(value, number) and value >= 0 else None


@declarations.register(
    "opacity",
)
def parse_factor(prop, value) -> Number | None:
    return value if isinstance(value, number) and 0 <= value <= 1 else None


@declarations.register("font-family")
def parse_string(prop, value) -> str | None:
    if isinstance(value, str):
        return value
    elif value:
        return " ".join(str(v) for v in value)
    return None


@declarations.register("font-size")
def parse_font_size(prop, value) -> None | int | float | str:
    if isinstance(value, number) and value > 0:
        return value
    return value if isinstance(value, str) and value in FONT_SIZE_VALUES else None


@declarations.register("padding")
def parse_padding(prop, value) -> Padding | None:
    if isinstance(value, number):
        return (value, value, value, value)
    n = len(value)
    if not n or any(not isinstance(v, number) for v in value):
        return None

    return (
        value[0],
        value[1],
        value[2] if n > 2 else value[0],
        value[3] if n > 3 else value[1],
    )


@declarations.register("dash-style")
def parse_sequence_numbers(
    prop, value: Number | Sequence[Number]
) -> Sequence[Number] | None:
    if value == 0:
        return ()
    if isinstance(value, number):
        return (value, value)
    elif all(isinstance(v, int | float) for v in value):
        return value
    return None


enum_styles: dict[str, dict[str, object]] = {
    "font-style": {e.value: e for e in FontStyle},
    "font-weight": {e.value: e for e in FontWeight},
    "justify-content": {e.value: e for e in JustifyContent},
    "text-decoration": {e.value: e for e in TextDecoration},
    "text-align": {e.value: e for e in TextAlign},
    "vertical-align": {e.value: e for e in VerticalAlign},
    "white-space": {e.value: e for e in WhiteSpace},
}


@declarations.register(*enum_styles.keys())
def parse_enum(prop, value):
    return enum_styles[prop].get(value)


@declarations.register("line-style")
def parse_line_style(prop, value) -> float:
    if value == "sloppy":
        return 0.5
    elif isinstance(value, tuple) and len(value) == 2:
        style, factor = value
        if style == "sloppy":
            if not isinstance(factor, number):
                return 0.5
            if factor < -2.0:
                return -2.0
            elif factor > 2.0:
                return 2.0
            else:
                return float(factor)
    # "normal" value:
    return 0.0


@declarations.register(
    "content",
)
def parse_content(prop, value) -> str:
    return value or ""



from gaphor.diagram.painter import DiagramTypePainter
from gaphor.SysML.diagramframe import DiagramFrameItem, diagram_label
from gaphor.SysML.sysml import SysMLDiagram
from gaphor.UML.painter import UMLDiagramTypePainter


@DiagramTypePainter.register(SysMLDiagram)  # type: ignore[attr-defined]
class SysMLDiagramTypePainter(UMLDiagramTypePainter):
    def label(self):
        return diagram_label(self.diagram)

    def paint(self, items, cr):
        if any(isinstance(p, DiagramFrameItem) for p in self.diagram.ownedPresentation):
            return

        super().paint(items, cr)



from gaphor.C4Model import c4model
from gaphor.core import Transaction
from gaphor.diagram.propertypages import (
    PropertyPageBase,
    PropertyPages,
    handler_blocking,
    new_resource_builder,
    unsubscribe_all_on_destroy,
)

new_builder = new_resource_builder("gaphor.C4Model")


@PropertyPages.register(c4model.Container)
@PropertyPages.register(c4model.Person)
class DescriptionPropertyPage(PropertyPageBase):
    order = 14

    def __init__(self, subject: c4model.Container | c4model.Person, event_manager):
        super().__init__()
        self.subject = subject
        self.event_manager = event_manager
        self.watcher = subject.watcher()

    def construct(self):
        builder = new_builder(
            "description-editor",
            "description-text-buffer",
        )
        subject = self.subject

        description = builder.get_object("description")

        buffer = builder.get_object("description-text-buffer")
        if subject.description:
            buffer.set_text(subject.description)

        @handler_blocking(buffer, "changed", self._on_description_changed)
        def text_handler(event):
            if not description.props.has_focus:
                buffer.set_text(event.new_value or "")

        self.watcher.watch("description", text_handler)

        return unsubscribe_all_on_destroy(
            builder.get_object("description-editor"), self.watcher
        )

    def _on_description_changed(self, buffer):
        with Transaction(self.event_manager, context="editing"):
            self.subject.description = buffer.get_text(
                buffer.get_start_iter(), buffer.get_end_iter(), False
            )


@PropertyPages.register(c4model.Container)
@PropertyPages.register(c4model.Dependency)
class TechnologyPropertyPage(PropertyPageBase):
    order = 15

    def __init__(self, subject: c4model.Container | c4model.Dependency, event_manager):
        super().__init__()
        self.subject = subject
        self.event_manager = event_manager
        self.watcher = subject.watcher()

    def construct(self):
        builder = new_builder(
            "technology-editor",
        )
        subject = self.subject

        technology = builder.get_object("technology")
        technology.set_text(subject.technology or "")

        @handler_blocking(technology, "changed", self._on_technology_changed)
        def text_handler(event):
            if (
                event.element is subject
                and (event.new_value or "") != technology.get_text()
            ):
                technology.set_text(event.new_value or "")

        self.watcher.watch("technology", text_handler)

        return unsubscribe_all_on_destroy(
            builder.get_object("technology-editor"), self.watcher
        )

    def _on_technology_changed(self, entry):
        with Transaction(self.event_manager, context="editing"):
            self.subject.technology = entry.get_text()



"""Stereotype property page."""

from unicodedata import normalize

from gi.repository import Gdk, Gio, GLib, GObject, Gtk

from gaphor import UML
from gaphor.core.eventmanager import EventManager
from gaphor.core.modeling import Presentation
from gaphor.diagram.propertypages import PropertyPageBase, PropertyPages
from gaphor.i18n import gettext, translated_ui_string
from gaphor.transaction import Transaction
from gaphor.UML.profiles.metaclasspropertypage import new_builder
from gaphor.UML.propertypages import text_field_handlers


@PropertyPages.register(UML.Element)
class StereotypePage(PropertyPageBase):
    order = 40

    def __init__(self, subject, event_manager):
        self.subject = subject
        self.event_manager = event_manager

    def construct(self):
        subject = self.subject
        if isinstance(subject, Presentation):
            subject = subject.subject

        if not subject:
            return None

        stereotypes = UML.recipes.get_stereotypes(subject)
        if not stereotypes:
            return None

        builder = new_builder(
            "stereotypes-editor",
            signals={
                "stereotype-key-pressed": (stereotype_key_handler,),
            },
        )

        stereotype_list = builder.get_object("stereotype-list")
        model = stereotype_model(subject, self.event_manager)

        stereotype_set_model_with_interaction(stereotype_list, model)

        return builder.get_object("stereotypes-editor")


@PropertyPages.register(Presentation)
class ShowStereotypePage(PropertyPageBase):
    order = 41

    def __init__(self, item, event_manager):
        self.item = item
        self.event_manager = event_manager

    def construct(self):
        subject = self.item.subject
        if not subject or not hasattr(self.item, "show_stereotypes"):
            return None

        stereotypes = UML.recipes.get_stereotypes(subject)
        if not stereotypes:
            return None

        builder = new_builder(
            "show-stereotypes-editor",
            signals={
                "show-stereotypes-changed": (self._on_show_stereotypes_change,),
            },
        )

        show_stereotypes = builder.get_object("show-stereotypes")
        show_stereotypes.set_active(self.item.show_stereotypes)

        return builder.get_object("show-stereotypes-editor")

    def _on_show_stereotypes_change(self, button, gparam):
        with Transaction(self.event_manager):
            self.item.show_stereotypes = button.get_active()


def stereotype_set_model_with_interaction(stereotype_list, model):
    def on_stereotype_toggled(cell):
        button = cell.get_child().get_first_child()
        view = cell.get_item()
        view.applied = button.get_active()
        update_stereotype_model(model)

    for column, factory in zip(
        stereotype_list.get_columns(),
        [
            name_list_item_factory(
                signal_handlers={
                    "on_toggled": (on_stereotype_toggled,),
                },
            ),
            value_list_item_factory(
                signal_handlers={
                    "stereotype-click-pressed": stereotype_click_handler,
                    **text_field_handlers("slot_value"),
                },
            ),
        ],
        strict=False,
    ):
        column.set_factory(factory)

    selection = Gtk.SingleSelection.new(model)
    stereotype_list.set_model(selection)


class StereotypeView(GObject.Object):
    def __init__(
        self,
        target: UML.Element,
        stereotype: UML.Stereotype,
        attr: UML.Property | None,
        event_manager: EventManager,
    ):
        super().__init__()
        self.stereotype = stereotype
        self.target = target
        self.attr = attr
        self.event_manager = event_manager

    @property
    def instance(self):
        return stereotype_application(self.stereotype, self.target)

    @property
    def slot(self):
        instance = self.instance
        attr = self.attr
        return (
            next(
                (slot for slot in instance.slot if slot.definingFeature is attr),
                None,
            )
            if instance
            else None
        )

    editing = GObject.Property(type=bool, default=False)

    @GObject.Property(type=str, default="", flags=GObject.ParamFlags.READABLE)
    def name(self):
        return (self.attr.name if self.attr else self.stereotype.name) or ""

    @GObject.Property(type=bool, default=False)
    def applied(self):
        return bool(self.instance)

    @applied.setter  # type: ignore[no-redef]
    def applied(self, value):
        instance = self.instance
        with Transaction(self.event_manager):
            if value and not instance:
                UML.recipes.apply_stereotype(self.target, self.stereotype)
            elif not value and instance:
                UML.recipes.remove_stereotype(self.target, self.stereotype)

    @GObject.Property(type=bool, default=False)
    def checkbox_visible(self):
        return not bool(self.attr)

    @GObject.Property(type=str)
    def placeholder_text(self):
        return gettext("New Slot Value…") if self.attr and self.instance else ""

    @GObject.Property(type=bool, default=False)
    def can_edit(self):
        return bool(self.attr and self.instance)

    @GObject.Property(type=GObject.TYPE_STRV)
    def indent_classes(self):
        return ["stereotype-value"] if self.attr else []

    @GObject.Property(type=bool, default=True)
    def sensitive(self):
        return not self.attr or self.instance

    @GObject.Property(type=str)
    def slot_value(self):
        slot = self.slot
        return UML.recipes.get_slot_value(slot) if slot else ""

    @slot_value.setter  # type: ignore[no-redef]
    def slot_value(self, value):
        slot = self.slot
        with Transaction(self.event_manager):
            if value:
                if slot is None:
                    slot = UML.recipes.add_slot(self.instance, self.attr)
                UML.recipes.set_slot_value(slot, value)
            elif slot:
                del self.instance.slot[slot]


def stereotype_model(subject: UML.Element, event_manager: EventManager):
    model = Gio.ListStore.new(StereotypeView.__gtype__)
    stereotypes = UML.recipes.get_stereotypes(subject)

    for st in sorted(
        stereotypes, key=lambda s: normalize("NFC", s.name or "").casefold()
    ):
        model.append(StereotypeView(subject, st, None, event_manager))

        for attr in all_attributes(st):
            model.append(StereotypeView(subject, st, attr, event_manager))

    return model


def update_stereotype_model(store: Gio.ListStore):
    for view in store:
        view.notify("placeholder_text")
        view.notify("can_edit")
        view.notify("slot_value")
        view.notify("sensitive")


def all_attributes(stereotype, seen=None):
    if seen is None:
        seen = {stereotype}

    for super_type in stereotype.generalization[:].general[:]:
        if super_type not in seen:
            seen.add(super_type)
            yield from all_attributes(super_type)
    yield from (attr for attr in stereotype.ownedAttribute if not attr.association)


def stereotype_application(st, subject) -> UML.InstanceSpecification | None:
    return next(
        (applied for applied in subject.appliedStereotype if st in applied.classifier),
        None,
    )


def name_list_item_factory(signal_handlers=None):
    ui_string = translated_ui_string("gaphor.UML.profiles", "stereotype-name-cell.ui")

    return Gtk.BuilderListItemFactory.new_from_bytes(
        Gtk.Builder.BuilderScope(signal_handlers),
        GLib.Bytes.new(ui_string.encode("utf-8")),
    )


def value_list_item_factory(signal_handlers=None):
    ui_string = translated_ui_string("gaphor.UML.profiles", "stereotype-value-cell.ui")

    return Gtk.BuilderListItemFactory.new_from_bytes(
        Gtk.Builder.BuilderScope(signal_handlers),
        GLib.Bytes.new(ui_string.encode("utf-8")),
    )


def stereotype_click_handler(ctrl, n_press, x, y):
    if n_press == 2:
        cell = ctrl.get_widget()
        cell.editing = True


def stereotype_key_handler(ctrl, keyval, _keycode, state):
    list_view = ctrl.get_widget()
    selection = list_view.get_model()
    item = selection.get_selected_item()

    # action: selection.rename
    if keyval in (Gdk.KEY_F2,):
        item.editing = True
        return True

    # action: selection.delete
    if keyval in (Gdk.KEY_Delete, Gdk.KEY_BackSpace) and not state & (
        Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK
    ):
        item.slot_value = ""
        return True

    return False



"""The action definition for the RAAML toolbox."""

from gaphor.diagram.diagramtoolbox import (
    DiagramType,
    DiagramTypes,
    ElementCreateInfo,
    ToolboxDefinition,
)
from gaphor.i18n import i18nize
from gaphor.RAAML.fta.ftatoolbox import fta
from gaphor.RAAML.raaml import (
    FTADiagram,
    Hazard,
    Loss,
    Situation,
    STPADiagram,
    TopEvent,
)
from gaphor.RAAML.stpa.stpatoolbox import stpa
from gaphor.UML.general.generaltoolbox import general_tools
from gaphor.UML.uml import Diagram, Package

raaml_toolbox_actions: ToolboxDefinition = (
    general_tools,
    fta,
    stpa,
)

raaml_diagram_types: DiagramTypes = (
    DiagramType(FTADiagram, i18nize("FTA Diagram"), (fta,)),
    DiagramType(STPADiagram, i18nize("STPA Diagram"), (stpa,)),
    DiagramType(Diagram, i18nize("Generic Diagram"), ()),
)

raaml_element_types = (
    ElementCreateInfo("package", i18nize("Package"), Package, (Package,)),
    ElementCreateInfo("topevent", i18nize("Top Event"), TopEvent, (Package,)),
    ElementCreateInfo("loss", i18nize("Loss"), Loss, (Package,)),
    ElementCreateInfo("hazard", i18nize("Hazard"), Hazard, (Package,)),
    ElementCreateInfo("situation", i18nize("Situation"), Situation, (Package,)),
)



import os
import sys
from pathlib import Path


def gi_init():
    """Workaround for https://gitlab.gnome.org/GNOME/pygobject/-/issues/545."""
    if sys.platform != "win32":
        import gi

        return gi
    env_path = os.environ.get("PATH", "").split(os.pathsep)
    if first_gtk_path := next(
        filter(
            lambda path: path is not None
            and Path.is_file(Path(path) / "girepository-2.0-0.dll"),
            env_path,
        ),
        None,
    ):
        with os.add_dll_directory(first_gtk_path):
            import gi

            return gi
    return None



"""Component item."""

from gaphor import UML
from gaphor.core.modeling.properties import attribute
from gaphor.diagram.presentation import Classified, ElementPresentation
from gaphor.diagram.shapes import Box, cairo_state, draw_border
from gaphor.diagram.support import represents
from gaphor.UML.classes.stereotype import stereotype_compartments, stereotype_watches
from gaphor.UML.compartments import name_compartment


@represents(UML.Component)
class ComponentItem(Classified, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)

        self.watch("show_stereotypes", self.update_shapes)
        self.watch("subject[NamedElement].name")
        self.watch("subject[NamedElement].namespace.name")
        self.watch("subject[Classifier].useCase", self.update_shapes)
        self.watch("children", self.update_shapes)
        stereotype_watches(self)

    show_stereotypes: attribute[int] = attribute("show_stereotypes", int)

    def update_shapes(self, event=None):
        self.shape = Box(
            name_compartment(self, draw_icon=draw_component_icon),
            *(self.show_stereotypes and stereotype_compartments(self.subject) or []),
            draw=draw_border,
        )


def draw_component_icon(box, context, bounding_box):
    with cairo_state(context.cairo) as cr:
        stroke_color = context.style["color"]

        cr.set_line_width(1.0)
        icon_height = 20
        icon_width = 16
        icon_margin_x = 6
        ix = bounding_box.width - icon_margin_x - icon_width
        icon_margin_y = 12

        iy = icon_margin_y

        bar_padding = 4
        bx = ix - bar_padding
        bar_upper_y = iy + bar_padding
        bar_lower_y = iy + bar_padding * 3

        bar_width = 12
        bar_height = 4
        cr.set_source_rgba(*stroke_color)

        cr.rectangle(bx, bar_lower_y, bar_width, bar_height)
        cr.rectangle(bx, bar_upper_y, bar_width, bar_height)

        cr.move_to(ix, iy + bar_padding)
        cr.line_to(ix, iy)
        cr.line_to(ix + icon_width, iy)
        cr.line_to(ix + icon_width, iy + icon_height)
        cr.line_to(ix, iy + icon_height)
        cr.line_to(ix, iy + icon_height - bar_padding)
        cr.move_to(ix, iy + bar_padding * 3)
        cr.line_to(ix, iy + bar_padding * 2)

        cr.stroke()



"""Connect comments."""

import logging

from gaphor.diagram.connectors import BaseConnector, Connector
from gaphor.diagram.presentation import ElementPresentation, LinePresentation
from gaphor.UML import Comment, Element
from gaphor.UML.general.comment import CommentItem, CommentLineItem

logger = logging.getLogger(__name__)


@Connector.register(CommentItem, CommentLineItem)
@Connector.register(ElementPresentation, CommentLineItem)
class CommentLineElementConnect(BaseConnector):
    """Connect a comment line to any element item."""

    element: CommentItem | ElementPresentation
    line: CommentLineItem

    def allow(self, handle, port):
        """In addition to the normal check, both line ends may not be connected
        to the same element.

        Same goes for subjects. One of the ends should be connected to a
        Comment element.
        """
        opposite = self.line.opposite(handle)
        connected_to = self.get_connected(opposite)
        element = self.element

        if connected_to is element:
            return None

        # Same goes for subjects:
        if (
            connected_to
            and not connected_to.subject
            and not element.subject
            and connected_to.subject is element.subject
        ):
            return None

        # One end should be connected to a CommentItem:
        cls = CommentItem
        glue_ok = isinstance(connected_to, cls) ^ isinstance(self.element, cls)
        if connected_to and not glue_ok:
            return None

        # Do not allow links between the comment and the element
        if (
            connected_to
            and element
            and (
                (
                    isinstance(connected_to.subject, Comment)
                    and self.element.subject in connected_to.subject.annotatedElement
                )
                or (
                    isinstance(self.element.subject, Comment)
                    and connected_to.subject in self.element.subject.annotatedElement
                )
            )
        ):
            return None

        return super().allow(handle, port)

    def connect(self, handle, port):
        if super().connect(handle, port):
            opposite = self.line.opposite(handle)
            connected_to = self.get_connected(opposite)
            if connected_to and connected_to.subject:
                if isinstance(connected_to.subject, Comment):
                    connected_to.subject.annotatedElement = self.element.subject
                else:
                    assert isinstance(self.element.subject, Comment)
                    assert isinstance(connected_to.subject, Element)
                    self.element.subject.annotatedElement = connected_to.subject

    def disconnect(self, handle):
        opposite = self.line.opposite(handle)
        oct = self.get_connected(opposite)
        hct = self.get_connected(handle)

        if hct and oct:
            logger.debug(f"Disconnecting {hct} and {oct}")
            try:
                if hct.subject:
                    if isinstance(oct.subject, Comment):
                        assert isinstance(hct.subject, Element)
                        del oct.subject.annotatedElement[hct.subject]
                    elif oct.subject:
                        assert isinstance(hct.subject, Comment)
                        assert isinstance(oct.subject, Element)
                        del hct.subject.annotatedElement[oct.subject]
            except ValueError:
                logger.debug(
                    "Invoked CommentLineElementConnect.disconnect() for nonexistent relationship"
                )

        super().disconnect(handle)


@Connector.register(LinePresentation, CommentLineItem)
class CommentLineLineConnect(BaseConnector):
    """Connect a comment line to any diagram line."""

    element: LinePresentation
    line: CommentLineItem

    def allow(self, handle, port):
        """In addition to the normal check, both line ends may not be connected
        to the same element.

        Same goes for subjects. One of the ends should be connected to a
        Comment element.
        """
        opposite = self.line.opposite(handle)
        element = self.element
        connected_to = self.get_connected(opposite)

        # do not connect to the same item nor connect to other comment line
        if connected_to is element or isinstance(element, CommentLineItem):
            return None

        # One end should be connected to a CommentItem:
        cls = CommentItem
        glue_ok = isinstance(connected_to, cls) ^ isinstance(self.element, cls)
        return None if connected_to and not glue_ok else super().allow(handle, port)

    def connect(self, handle, port):
        if super().connect(handle, port):
            opposite = self.line.opposite(handle)
            c = self.get_connected(opposite)
            if c and self.element.subject:
                if isinstance(c.subject, Comment):
                    c.subject.annotatedElement = self.element.subject
                else:
                    assert isinstance(self.element.subject, Comment)
                    assert isinstance(c.subject, Element)
                    self.element.subject.annotatedElement = c.subject

    def disconnect(self, handle):
        c1 = self.get_connected(handle)
        opposite = self.line.opposite(handle)
        c2 = self.get_connected(opposite)
        if c1 and c2:
            if (
                isinstance(c1.subject, Comment)
                and isinstance(c2.subject, Element)
                and c2.subject in c1.subject.annotatedElement
            ):
                del c1.subject.annotatedElement[c2.subject]
            elif (
                isinstance(c1.subject, Element)
                and isinstance(c2.subject, Comment)
                and c1.subject in c2.subject.annotatedElement
            ):
                del c2.subject.annotatedElement[c1.subject]
        super().disconnect(handle)



"""Connector adapters.

To register connectors implemented in this module, it is imported in
gaphor.adapter package.
"""

from __future__ import annotations

import functools
import itertools
from collections.abc import Iterator
from typing import Protocol, TypeVar

from gaphas.connections import Connection
from gaphas.connector import ConnectionSink, Handle, Port
from gaphas.connector import Connector as ConnectorAspect
from generic.multidispatch import FunctionDispatcher, multidispatch

from gaphor.core.modeling import Base, Diagram, Presentation
from gaphor.core.modeling.event import RevertibleEvent
from gaphor.core.modeling.properties import association, redefine, relation
from gaphor.diagram.copypaste import copy, paste
from gaphor.diagram.presentation import ElementPresentation, LinePresentation
from gaphor.diagram.support import get_diagram_item_metadata, get_model_element

T = TypeVar("T", bound=Base)


class ConnectorProtocol(Protocol):
    def __init__(
        self,
        element: Presentation,
        line: Presentation,
    ) -> None: ...

    def allow(self, handle: Handle, port: Port) -> bool: ...

    def connect(self, handle: Handle, port: Port) -> bool: ...

    def disconnect(self, handle: Handle) -> None: ...


class BaseConnector:
    """Connection adapter for Gaphor diagram items.

    Line item ``line`` connects with a handle to a connectable item ``element``.

    :param line: connecting item
    :param element: connectable item

    By convention the adapters are registered by (element, line) -- in that order.
    """

    def __init__(
        self,
        element: Presentation[Base],
        line: Presentation[Base],
    ) -> None:
        self.element = element
        self.line = line

        # Try to find the diagram. Any diagram can be None, when called while an element is unlinking
        assert (
            element.diagram is line.diagram
            or element.diagram is None
            or line.diagram is None
        )
        self.diagram: Diagram = element.diagram or line.diagram
        assert self.diagram

    def get_connected(self, handle: Handle) -> Presentation[Base] | None:
        """Get item connected to a handle."""
        if cinfo := self.diagram.connections.get_connection(handle):
            return cinfo.connected  # type: ignore[no-any-return] # noqa: F723
        return None

    def allow(self, handle: Handle, port: Port) -> bool:
        """Determine if items can be connected.

        Is the connection allowed at all (during mouse movement for example)?

        Returns `True` if connection is allowed.
        """
        return bool(self.element.diagram) and self.element.diagram is self.line.diagram

    def connect(self, handle: Handle, port: Port) -> bool:
        """Connect to an element.

        Establish a connection between element and line. Also takes care of
        disconnects, if required (e.g. 1:1 relationships).

        Note that at this point the line may be
        connected to some other, or the same element. The connection at model
        level also still exists.

        Returns `True` if a connection is established.
        """
        return True

    def disconnect(self, handle: Handle) -> None:
        """Disconnect model level connections.

        Break connection, called when dropping a handle on a
        point where it can not connect.
        """


class NoConnector:
    def __init__(
        self,
        element,
        line,
    ) -> None:
        pass

    def allow(self, handle: Handle, port: Port) -> bool:
        return False

    def connect(self, handle: Handle, port: Port) -> bool:
        return False

    def disconnect(self, handle: Handle) -> None:
        pass


# Work around issue https://github.com/python/mypy/issues/3135 (Class decorators are not type checked)
# This definition, along with the ignore below, seems to fix the behaviour for mypy at least.
Connector: FunctionDispatcher[type[ConnectorProtocol]] = multidispatch(object, object)(
    NoConnector
)


@functools.lru_cache
def can_connect(parent, element_type) -> bool:
    parent_type = type(parent)
    get_registration = Connector.registry.get_registration
    for t1, t2 in itertools.product(parent_type.__mro__, element_type.__mro__):
        if r := get_registration(t1, t2):
            return r is not NoConnector
    return False


class RelationshipConnect(BaseConnector):
    """Base class for relationship connections, such as associations,
    dependencies and implementations.

    Unary relationships are allowed to connect both ends to the same element.

    This class introduces a new method: `relationship()`, which is used to
    find an existing relationship in the model that does not yet exist
    on the diagram.
    """

    element: Presentation
    line: LinePresentation[Base]

    def __init__(self, element: Presentation[Base], line: Presentation[Base]) -> None:
        super().__init__(element, line)
        self.copy_buffer = dict(copy(line.subject)) if line.subject else {}

    def relationship(
        self, required_type: type[T], head: relation, tail: relation
    ) -> T | None:
        """Find an existing relationship in the model that meets the required
        type and is connected to the same model element the head and tail of
        the line are connected to.

        type - the type of relationship we're looking for
        head - tuple (association name on the line, association name on the element)
        tail - tuple (association name on the line, association name on the element)
        """
        assert isinstance(head, association | redefine), f"head is {head}"
        assert isinstance(tail, association | redefine), f"tail is {tail}"

        line = self.line

        line_head = self.get_connected(line.head)
        line_tail = self.get_connected(line.tail)
        assert line_head
        assert line_tail
        head_subject = line_head.subject
        tail_subject = line_tail.subject

        # First check if the right subject is already connected:
        if (
            line.subject
            and head.get(line.subject) is head_subject
            and tail.get(line.subject) is tail_subject
        ):
            return line.subject  # type: ignore[return-value]

        # Try to find a relationship, that is already created, but not
        # yet displayed in the diagram on the tail side, since tail should
        # have a reference to head at least.
        if not (head_subject and tail.opposite):
            return None

        gen: Base
        for gen in getattr(tail_subject, tail.opposite):
            if not isinstance(gen, required_type):
                continue

            gen_head = head.get(gen)
            try:
                if head_subject not in gen_head:
                    continue
            except TypeError:
                if gen_head is not head_subject:
                    continue

            # Check for this entry on line.diagram
            item: ElementPresentation | LinePresentation
            for item in gen.presentation:
                # Allow line to be returned. Avoids strange
                # behaviour during loading
                if item.diagram is line.diagram and item is not line:
                    break
            else:
                return gen
        return None

    def new_relation(self, type: type[T]) -> T:
        return self.line.model.create(type)

    def new_relation_from_copy(self, type: type[T]) -> T | None:
        if not self.copy_buffer:
            return None

        for e in paste_model(self.copy_buffer, self.diagram):
            if isinstance(e, type):
                return e

        raise AssertionError(
            f"Copied elements, but no {type} found ({self.copy_buffer})"
        )

    def relationship_or_new(self, type: type[T], head: relation, tail: relation) -> T:
        """Like relation(), but create a new instance if none was found."""
        relation = self.relationship(type, head, tail)
        if relation:
            return relation

        line = self.line
        relation = self.new_relation_from_copy(type) or self.new_relation(type)

        assert isinstance(relation, type)
        line_head = self.get_connected(line.head)
        line_tail = self.get_connected(line.tail)

        assert line_head
        assert line_tail
        head.set(relation, line_head.subject)
        tail.set(relation, line_tail.subject)

        # TODO uncomment the following after Dependency is made a subclass of PackageableElement
        # owning_package = line_tail.subject.owningPackage
        # relation.owningPackage = owning_package

        assert isinstance(relation, type)
        return relation

    def connect_connected_items(self, connections: None = None) -> None:
        """Cause items connected to ``line`` to reconnect, allowing them to
        establish or destroy relationships at model level."""
        line = self.line
        diagram = self.diagram

        # First make sure coordinates match
        diagram.connections.solver.solve()
        for cinfo in connections or diagram.connections.get_connections(connected=line):
            if line is cinfo.connected:
                continue
            adapter = Connector(line, cinfo.connected)
            assert adapter, f"No element to connect {line} and {cinfo.connected}"
            adapter.connect(cinfo.handle, cinfo.port)

    def disconnect_connected_items(self) -> list[Connection]:
        """Cause items connected to @line to be disconnected. This is necessary
        if the subject of the @line is to be removed.

        Returns a list of (item, handle) pairs that were connected (this
        list can be used to connect items again with
        connect_connected_items()).
        """
        line = self.line
        diagram = self.diagram

        # First make sure coordinates match
        diagram.connections.solver.solve()
        connections = list(diagram.connections.get_connections(connected=line))
        for cinfo in connections:
            adapter = Connector(cinfo.item, cinfo.connected)
            adapter.disconnect(cinfo.handle)
        return connections

    def connect_subject(self, handle: Handle) -> bool:
        """Establish the relationship at model level."""
        raise NotImplementedError("Implement connect_subject() in a subclass")

    def disconnect_subject(self, handle: Handle) -> None:
        """Disconnect the diagram item from its model element."""
        del self.line.subject

    def connect(self, handle: Handle, port: Port) -> bool:
        """Connect the items to each other.

        The model level relationship is created by create_subject()
        """
        if not super().connect(handle, port):
            return False
        opposite = self.line.opposite(handle)
        if self.get_connected(opposite):
            self.connect_subject(handle)
            line = self.line
            if line.subject:
                self.connect_connected_items()
        return True

    def disconnect(self, handle: Handle) -> None:
        """Disconnect model element."""
        line = self.line
        opposite = line.opposite(handle)
        oct = self.get_connected(opposite)
        hct = self.get_connected(handle)

        if hct and oct:
            # Both sides of line are connected => disconnect
            self.disconnect_connected_items()
            self.disconnect_subject(handle)

        super().disconnect(handle)


class DirectionalRelationshipConnect(RelationshipConnect):
    """Base for relationship connections between unique elements."""

    def allow(self, handle: Handle, port: Port) -> bool:
        """In addition to the normal check, both relationship ends may not be
        connected to the same element.

        Same goes for subjects.
        """
        opposite = self.line.opposite(handle)
        element = self.element
        connected_to = self.get_connected(opposite)

        # Element can not be a parent of itself.
        if connected_to is element:
            return False

        # Same goes for subjects:
        if (
            connected_to
            and not connected_to.subject
            and not element.subject
            and connected_to.subject is element.subject
        ):
            return False

        return super().allow(handle, port)


@Connector.register(ElementPresentation, LinePresentation)
class MetadataRelationConnect(DirectionalRelationshipConnect):
    """A generic relationship connector."""

    def allow(self, handle, port):
        if not super().allow(handle, port):
            return False

        element = self.element
        opposite_element = self.get_connected(self.line.opposite(handle))
        if metadata := get_diagram_item_metadata(type(self.line)):
            return (
                (
                    isinstance(element.subject, metadata["head"].type)
                    and (
                        not opposite_element
                        or isinstance(opposite_element.subject, metadata["tail"].type)
                    )
                )
                if handle is self.line.head
                else (
                    isinstance(element.subject, metadata["tail"].type)
                    and (
                        not opposite_element
                        or isinstance(opposite_element.subject, metadata["head"].type)
                    )
                )
            )

        return False

    def connect_subject(self, handle):
        subject_type = get_model_element(type(self.line))
        metadata = get_diagram_item_metadata(type(self.line))
        if not metadata:
            return False

        self.line.subject = self.relationship_or_new(
            subject_type, metadata["head"], metadata["tail"]
        )
        return True


def paste_model(copy_data, diagram) -> Iterator[Base]:
    new_elements: dict[str, Base] = {}

    def create(ref: str):
        if ref in new_elements:
            return new_elements[ref]

        if ref in copy_data:
            paster = paste(copy_data[ref], diagram, create)
            new_elements[ref] = next(paster)
            next(paster, None)
            return new_elements[ref]

    for old_id in copy_data:
        if old_id in new_elements:
            continue
        create(old_id)

    for element in new_elements.values():
        assert element
        element.postload()

    return iter(new_elements.values())


class ItemConnected(RevertibleEvent):
    def __init__(self, element, handle, connected, port):
        super().__init__(element)
        self.handle_index = element.handles().index(handle)
        self.connected_id = connected.id
        self.port_index = connected.ports().index(port)

    def revert(self, target):
        # Reverse only the diagram level connection.
        # Associations have their own handlers
        connections = target.diagram.connections
        handle = target.handles()[self.handle_index]

        connector = ConnectorAspect(target, handle, connections)
        connector.disconnect_handle()


class ItemDisconnected(RevertibleEvent):
    def __init__(self, element, handle, connected, port):
        super().__init__(element)
        self.handle_index = element.handles().index(handle)
        self.connected_id = connected.id
        self.port_index = connected.ports().index(port)

    def revert(self, target):
        # Reverse only the diagram level connection.
        # Associations have their own handlers
        handle = target.handles()[self.handle_index]
        connections = target.diagram.connections
        connected = target.diagram.lookup(self.connected_id)

        sink = ConnectionSink(connected)
        sink.port = connected.ports()[self.port_index]
        connector = ConnectorAspect(target, handle, connections)
        connector.connect_handle(sink)


class ItemTemporaryDisconnected(RevertibleEvent):
    def __init__(self, element, handle, connected, port):
        super().__init__(element)
        self.handle_index = element.handles().index(handle)
        self.connected_id = connected.id
        self.port_index = connected.ports().index(port)

    def revert(self, target):
        handle = target.handles()[self.handle_index]
        connections = target.diagram.connections
        connected = target.diagram.lookup(self.connected_id)

        sink = ConnectionSink(connected)
        sink.port = connected.ports()[self.port_index]
        connector = ConnectorAspect(target, handle, connections)
        connector.reconnect_handle(sink)


class ItemReconnected(RevertibleEvent):
    def __init__(self, element, handle, connected, port):
        super().__init__(element)
        self.handle_index = element.handles().index(handle)
        self.connected_id = connected.id
        self.port_index = connected.ports().index(port)

    def revert(self, target):
        connections = target.diagram.connections
        handle = target.handles()[self.handle_index]

        cinfo = connections.get_connection(handle)
        connections.solver.remove_constraint(cinfo.constraint)
        target.handle(
            ItemTemporaryDisconnected(target, handle, cinfo.connected, cinfo.port)
        )



"""Undo management for Gaphor.

Undoing and redoing actions is managed through the UndoManager.

An undo action should be a callable object (called with no arguments).

An undo action should return a callable object that acts as redo function.
If None is returned the undo action is considered to be the redo action as well.

NOTE: it would be nice to use actions in conjunction with functools.partial.
"""

from __future__ import annotations

import logging
from collections.abc import Callable
from dataclasses import dataclass

from gaphor.abc import ActionProvider, Service
from gaphor.core import event_handler
from gaphor.core.modeling.base import Base, RepositoryProtocol, swap_element_type
from gaphor.core.modeling.event import (
    AssociationAdded,
    AssociationDeleted,
    AssociationSet,
    AttributeUpdated,
    ElementCreated,
    ElementDeleted,
    ElementTypeUpdated,
    ModelFlushed,
    ModelReady,
    RevertibleEvent,
)
from gaphor.core.modeling.presentation import Presentation
from gaphor.core.modeling.properties import association as association_property
from gaphor.diagram.copypaste import deserialize, serialize
from gaphor.event import (
    ActionEnabled,
    TransactionBegin,
    TransactionClosed,
    TransactionCommit,
    TransactionRollback,
)
from gaphor.transaction import Transaction

logger = logging.getLogger(__name__)


class ActionStack:
    """A transaction.

    Every action that is added between a begin_transaction() and a
    commit_transaction() call is recorded in a transaction, so it can be
    played back when a transaction is executed. This executing a
    transaction has the effect of performing the actions recorded, which
    will typically undo actions performed by the user.
    """

    def __init__(self):
        self._actions: list[Callable[[], None]] = []

    def add(self, action):
        self._actions.append(action)

    def can_execute(self):
        return bool(self._actions)

    def execute(self):
        self._actions.reverse()

        for act in self._actions:
            logger.debug(act.__doc__)
            act()


class EditingStack(ActionStack):
    """A transaction for text editing.

    Editing stacks can add actions over multiple transactions.

    An editing action can be closed, which means no more actions can be added.
    """

    def __init__(self):
        super().__init__()
        self.closed = False

    def close(self):
        self.closed = True


@dataclass
class UndoManagerStateChanged:
    """Event class used to send state changes on the Undo Manager."""

    undo_manager: UndoManager


class _UndoManagerTransactionCommitted:
    """A transaction was committed.

    This is a subsequent event that triggers the actual commit.
    In doing so we allow other parties to handle their commit handlers first.
    """

    def __init__(self, context):
        self.context = context


class _UndoManagerTransactionRolledBack:
    """A transaction has to be rolled back."""

    def __init__(self, context):
        self.context = context


class NotInTransactionException(Exception):
    """Raised when changes occur outside a transaction."""


class UndoManager(Service, ActionProvider):
    """Simple transaction manager for Gaphor. This transaction manager supports
    nested transactions.

    The Undo manager sports an undo and a redo stack. Each stack
    contains a set of actions that can be executed, just by calling them
    (e.i action()) If something is returned by an action, that is
    considered the callable to be used to undo or redo the last
    performed action.

    Change events (attribute/association updates) are handled with priority
    by the undo manager. This is done, so that, if a transaction is rolled back,
    all changes that have been applied are rolled back. This prevents events from
    being delayed because of exceptions that occur in some other event handler.
    """

    def __init__(self, event_manager, element_factory):
        self.event_manager = event_manager
        self.element_factory: RepositoryProtocol = element_factory
        self._undo_stack: list[ActionStack] = []
        self._redo_stack: list[ActionStack] = []
        self._stack_depth = 20
        self._current_transaction = None

        event_manager.subscribe(self.ready)
        event_manager.subscribe(self.reset)
        event_manager.priority_subscribe(self.begin_transaction)
        event_manager.subscribe(self.commit_transaction)
        event_manager.subscribe(self.rollback_transaction)
        event_manager.subscribe(self.close_transaction)
        event_manager.subscribe(self._on_transaction_commit)
        event_manager.subscribe(self._on_transaction_rollback)

        event_manager.priority_subscribe(self.undo_reversible_event)
        event_manager.priority_subscribe(self.undo_create_element_event)
        event_manager.priority_subscribe(self.undo_delete_element_event)
        event_manager.priority_subscribe(self.undo_attribute_change_event)
        event_manager.priority_subscribe(self.undo_association_set_event)
        event_manager.priority_subscribe(self.undo_association_add_event)
        event_manager.priority_subscribe(self.undo_association_delete_event)
        event_manager.priority_subscribe(self.undo_element_type_updated_event)
        self._action_executed(state_changed=False)

    def shutdown(self):
        self.event_manager.unsubscribe(self.ready)
        self.event_manager.unsubscribe(self.reset)
        self.event_manager.unsubscribe(self.begin_transaction)
        self.event_manager.unsubscribe(self.commit_transaction)
        self.event_manager.unsubscribe(self.rollback_transaction)
        self.event_manager.unsubscribe(self.close_transaction)
        self.event_manager.unsubscribe(self._on_transaction_commit)
        self.event_manager.unsubscribe(self._on_transaction_rollback)

        self.event_manager.unsubscribe(self.undo_reversible_event)
        self.event_manager.unsubscribe(self.undo_create_element_event)
        self.event_manager.unsubscribe(self.undo_delete_element_event)
        self.event_manager.unsubscribe(self.undo_attribute_change_event)
        self.event_manager.unsubscribe(self.undo_association_set_event)
        self.event_manager.unsubscribe(self.undo_association_add_event)
        self.event_manager.unsubscribe(self.undo_association_delete_event)
        self.event_manager.unsubscribe(self.undo_element_type_updated_event)

    def clear_undo_stack(self):
        del self._undo_stack[:]

    def clear_redo_stack(self):
        del self._redo_stack[:]

    @event_handler(ModelReady)
    def ready(self, _event=None):
        self._action_executed(state_changed=False)

    @event_handler(ModelFlushed)
    def reset(self, _event=None):
        assert not self._current_transaction

        self.clear_redo_stack()
        self.clear_undo_stack()
        self._action_executed(state_changed=False)

    @event_handler(TransactionBegin)
    def begin_transaction(self, event=None):
        """Add an action to the current transaction."""
        assert not self._current_transaction
        if event and event.context == "editing":
            if (
                self._undo_stack
                and isinstance(self._undo_stack[-1], EditingStack)
                and not self._undo_stack[-1].closed
            ):
                # "steal" the last transaction for amending
                self._current_transaction = self._undo_stack.pop()
            else:
                self._current_transaction = EditingStack()
        else:
            self._current_transaction = ActionStack()

    def add_undo_action(self, action):
        """Add an action to undo."""
        if self._current_transaction:
            self._current_transaction.add(action)
            self._action_executed()
        else:
            with Transaction(self.event_manager, context="rollback"):
                action()

            raise NotInTransactionException(
                f"Updating state outside of a transaction: {action.__doc__}."
            )

    @event_handler(TransactionCommit)
    def _on_transaction_commit(self, event: TransactionCommit):
        self.event_manager.handle(_UndoManagerTransactionCommitted(event.context))

    @event_handler(_UndoManagerTransactionCommitted)
    def commit_transaction(self, event: _UndoManagerTransactionCommitted | None = None):
        assert self._current_transaction

        if event is None:
            event = _UndoManagerTransactionCommitted(None)

        current_transaction = self._current_transaction
        self._current_transaction = None

        if event.context != "rollback" and current_transaction.can_execute():
            if event.context == "undo":
                self._redo_stack.append(current_transaction)
            else:
                if event.context != "redo":
                    self.clear_redo_stack()
                self._undo_stack.append(current_transaction)
                while len(self._undo_stack) > self._stack_depth:
                    del self._undo_stack[0]

            self._action_executed()

    @event_handler(TransactionRollback)
    def _on_transaction_rollback(self, event: TransactionRollback):
        self.event_manager.handle(_UndoManagerTransactionRolledBack(event.context))

    @event_handler(_UndoManagerTransactionRolledBack)
    def rollback_transaction(self, event=None):
        """Roll back the transaction we're in."""
        assert self._current_transaction

        if event and event.context in ("undo", "redo", "rollback"):
            logger.error(
                "Already performing %s, ignoring additional rollback events",
                event.context,
            )
            return

        erroneous_tx = self._current_transaction
        self._current_transaction = None
        with Transaction(self.event_manager, context="rollback"):
            try:
                erroneous_tx.execute()
            except Exception:
                logger.error("Could not rollback transaction", exc_info=True)
                raise

        self._action_executed()

    @event_handler(TransactionClosed)
    def close_transaction(self, _event=None):
        if self._current_transaction:
            logger.debug("Trying to close a transaction, while in a transaction")
            return

        if self._undo_stack:
            last_transaction = self._undo_stack[-1]
            if isinstance(last_transaction, EditingStack):
                last_transaction.close()

    def undo_transaction(self):
        if not self._undo_stack:
            return

        if self._current_transaction:
            logger.warning("Trying to undo a transaction, while in a transaction")
            self.commit_transaction()

        transaction = self._undo_stack.pop()
        with Transaction(self.event_manager, context="undo"):
            transaction.execute()

        self._action_executed()

    def redo_transaction(self):
        if not self._redo_stack:
            return

        assert not self._current_transaction

        transaction = self._redo_stack.pop()
        with Transaction(self.event_manager, context="redo"):
            transaction.execute()

        self._action_executed()

    def can_undo(self):
        return bool(self._current_transaction or self._undo_stack)

    def can_redo(self):
        return bool(self._redo_stack)

    def _action_executed(self, state_changed=True):
        self.event_manager.handle(ActionEnabled("text.undo", self.can_undo()))
        self.event_manager.handle(ActionEnabled("text.redo", self.can_redo()))
        if state_changed:
            self.event_manager.handle(UndoManagerStateChanged(self))

    def lookup(self, id: str) -> Base:
        if element := self.element_factory.lookup(id):
            return element
        else:
            raise ValueError(f"Element with id {id} not found in model")

    #
    # Undo Handlers
    #

    @event_handler(RevertibleEvent)
    def undo_reversible_event(self, event: RevertibleEvent):
        element_id = event.element.id

        def undo_reversible_event():
            element = self.lookup(element_id)
            event.revert(element)

        undo_reversible_event.__doc__ = (
            f"Reverse event {event.__class__.__name__} for element {event.element}."
        )

        self.add_undo_action(undo_reversible_event)

    @event_handler(ElementCreated)
    def undo_create_element_event(self, event: ElementCreated):
        element_id = event.element.id

        def undo_create_event():
            element = self.lookup(element_id)
            element.unlink()

        undo_create_event.__doc__ = f"Undo create element {event.element}."
        del event

        self.add_undo_action(undo_create_event)

    @event_handler(ElementDeleted)
    def undo_delete_element_event(self, event: ElementDeleted):
        element_type = type(event.element)
        element_id = event.element.id

        if isinstance(event.element, Presentation):
            diagram_id = event.diagram.id
            data = {}

            def save_func(name, value):
                data[name] = serialize(value)

            event.element.save(save_func)

            def undo_delete_event():
                diagram = self.lookup(diagram_id)
                element = diagram.create_as(element_type, element_id)  # type: ignore[attr-defined]

                for name, ser in data.items():
                    for value in deserialize(ser, lambda ref: None):
                        element.load(name, value)
        else:

            def undo_delete_event():
                self.element_factory.create_as(element_type, element_id)

        undo_delete_event.__doc__ = f"Recreate element {element_type} ({element_id})."
        del event

        self.add_undo_action(undo_delete_event)

    @event_handler(AttributeUpdated)
    def undo_attribute_change_event(self, event: AttributeUpdated):
        attribute = event.property
        element_id = event.element.id
        value = event.old_value

        def undo_attribute_change_event():
            element = self.lookup(element_id)
            attribute.set(element, value)

        undo_attribute_change_event.__doc__ = (
            f"Revert {event.element}.{attribute.name} to {value}."
        )
        del event

        self.add_undo_action(undo_attribute_change_event)

    @event_handler(AssociationSet)
    def undo_association_set_event(self, event: AssociationSet):
        association = event.property
        if type(association) is not association_property:
            return
        element_id = event.element.id
        value_id = event.old_value and event.old_value.id

        def undo_association_set_event():
            element = self.lookup(element_id)
            value = value_id and self.lookup(value_id)
            association.set(element, value, from_opposite=True)

        undo_association_set_event.__doc__ = (
            f"Revert {event.element}.{association.name} to {event.old_value}."
        )
        del event

        self.add_undo_action(undo_association_set_event)

    @event_handler(AssociationAdded)
    def undo_association_add_event(self, event: AssociationAdded):
        association = event.property
        if type(association) is not association_property:
            return
        element_id = event.element.id
        value_id = event.new_value.id

        def undo_association_add_event():
            element = self.lookup(element_id)
            value = self.lookup(value_id)
            association.delete(element, value, from_opposite=True)

        undo_association_add_event.__doc__ = (
            f"{event.element}.{association.name} delete {event.new_value}."
        )
        del event

        self.add_undo_action(undo_association_add_event)

    @event_handler(AssociationDeleted)
    def undo_association_delete_event(self, event: AssociationDeleted):
        association = event.property
        if type(association) is not association_property:
            return
        element_id = event.element.id
        value_id = event.old_value.id
        index = event.index

        def undo_association_delete_event():
            element = self.lookup(element_id)
            value = self.lookup(value_id)
            association.set(element, value, index=index, from_opposite=True)

        undo_association_delete_event.__doc__ = (
            f"{event.element}.{association.name} add {event.old_value}."
        )
        del event

        self.add_undo_action(undo_association_delete_event)

    @event_handler(ElementTypeUpdated)
    def undo_element_type_updated_event(self, event: ElementTypeUpdated):
        element_id = event.element.id
        old_class = event.old_class

        def undo_element_type_updated_event():
            element = self.lookup(element_id)
            swap_element_type(element, old_class)

        undo_element_type_updated_event.__doc__ = (
            f"{event.element} class is {old_class}."
        )
        del event

        self.add_undo_action(undo_element_type_updated_event)



"""Containment - A relationship that makes an item an ownedElement of another."""

from gaphor.core.modeling import DrawContext
from gaphor.diagram.presentation import LinePresentation
from gaphor.diagram.shapes import pi, stroke


class ContainmentItem(LinePresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)
        self.draw_head = draw_crossed_circle_head


def draw_crossed_circle_head(context: DrawContext):
    cr = context.cairo
    radius = 7
    cr.save()
    cr.move_to(0, 0)
    cr.set_dash((), 0)
    cr.arc(radius, 0.0, radius, 0.0, 2 * pi)
    cr.move_to(0, 0)
    cr.line_to(2 * radius, 0)
    cr.move_to(radius, radius)
    cr.line_to(radius, -radius)
    stroke(context, fill=False, dash=False)
    cr.restore()
    cr.move_to(0, 0)



from gaphor import UML
from gaphor.diagram.propertypages import (
    PropertyPageBase,
    PropertyPages,
    new_resource_builder,
)
from gaphor.transaction import Transaction
from gaphor.UML.interactions.interactionsconnect import get_lifeline
from gaphor.UML.interactions.message import MessageItem
from gaphor.UML.propertypages import list_of_classifiers

new_builder = new_resource_builder("gaphor.UML.interactions")


@PropertyPages.register(UML.Lifeline)
class LifelinePropertyPage(PropertyPageBase):
    order = 31

    def __init__(self, subject: UML.Lifeline, event_manager):
        super().__init__()
        self.subject = subject
        self.event_manager = event_manager

    def construct(self):
        if not self.subject:
            return

        builder = new_builder(
            "lifeline-editor",
        )

        dropdown = builder.get_object("element-type")
        model = list_of_classifiers(self.subject.model, UML.ConnectableElement)
        dropdown.set_model(model)

        if self.subject.represents:
            dropdown.set_selected(
                next(
                    n
                    for n, lv in enumerate(model)
                    if lv.value == self.subject.represents.id
                )
            )

        dropdown.connect("notify::selected", self._on_property_type_changed)

        return builder.get_object("lifeline-editor")

    def _on_property_type_changed(self, dropdown, _pspec):
        subject = self.subject
        with Transaction(self.event_manager):
            if id := dropdown.get_selected_item().value:
                element = subject.model.lookup(id)
                assert isinstance(element, UML.ConnectableElement)
                subject.represents = element
            else:
                del subject.represents


@PropertyPages.register(MessageItem)
class MessagePropertyPage(PropertyPageBase):
    """Property page for editing message items.

    When message is on communication diagram, then additional messages
    can be added. On sequence diagram sort of message can be changed.
    """

    order = 15

    MESSAGE_SORT = (
        "synchCall",
        "asynchCall",
        "asynchSignal",
        "reply",
        "createMessage",
        "deleteMessage",
    )

    def __init__(self, item, event_manager):
        self.item = item
        self.event_manager = event_manager

    def construct(self):
        item = self.item
        subject = item.subject

        if not subject or item.is_communication():
            return

        builder = new_builder(
            "message-editor",
            signals={"message-combo-changed": (self._on_message_sort_change,)},
        )

        lifeline = get_lifeline(item, item.tail)

        dropdown = builder.get_object("message-combo")

        # disallow connecting two delete messages to a lifeline
        if (
            lifeline
            and lifeline.is_destroyed
            and subject.messageSort != "deleteMessage"
        ):
            dropdown.get_model().remove(self.MESSAGE_SORT.index("deleteMessage"))

        dropdown.set_selected(self.MESSAGE_SORT.index(subject.messageSort))

        return builder.get_object("message-editor")

    def _on_message_sort_change(self, dropdown, _pspec):
        """Update message item's message sort information."""

        ms = self.MESSAGE_SORT[dropdown.get_selected()]

        item = self.item
        subject = item.subject
        lifeline = get_lifeline(item, item.tail)

        with Transaction(self.event_manager):
            # allow only one delete message to connect to lifeline's lifetime
            # destroyed status can be changed only by delete message itself
            if lifeline and (
                subject.messageSort == "deleteMessage" or not lifeline.is_destroyed
            ):
                is_destroyed = ms == "deleteMessage"
                lifeline.is_destroyed = is_destroyed
                lifeline.request_update()

            subject.messageSort = ms
            item.request_update()



"""Package diagram item."""

from gaphor import UML
from gaphor.diagram.presentation import (
    ElementPresentation,
    Named,
    text_name,
)
from gaphor.diagram.shapes import Box, cairo_state, stroke
from gaphor.diagram.support import represents
from gaphor.UML.compartments import text_from_package, text_stereotypes


@represents(UML.Package)
@represents(UML.Profile)
class PackageItem(Named, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=70, height=70)

        self.watch("children", self.update_shapes)
        self.watch("subject[NamedElement].name")
        self.watch("subject[NamedElement].namespace.name")
        self.watch("subject[Element].appliedStereotype.classifier.name")

    def update_shapes(self, event=None):
        self.shape = Box(
            text_stereotypes(
                self,
                lambda: (
                    [self.diagram.gettext("profile")]
                    if isinstance(self.subject, UML.Profile)
                    else []
                ),
            ),
            text_name(self),
            text_from_package(self),
            draw=draw_package,
        )


def draw_package(box, context, bounding_box):
    with cairo_state(context.cairo) as cr:
        o = 0.0
        h = bounding_box.height
        w = bounding_box.width
        x = 50
        y = 20
        cr.move_to(x, y)
        cr.line_to(x, o)
        cr.line_to(o, o)
        cr.line_to(o, h)
        cr.line_to(w, h)
        cr.line_to(w, y)
        cr.line_to(o, y)
        stroke(context, fill=True)



"""Support for actions in generic files."""

from __future__ import annotations

import asyncio
import inspect
import logging
from collections.abc import Callable, Iterator
from typing import Any, get_type_hints

log = logging.getLogger(__name__)


class action:
    """Decorator. Turns a regular function (/method) into a full blown Action
    class.

    >>> class A:
    ...     @action(name="my_action", label="my action")
    ...     def myaction(self):
    ...         print("action called")
    >>> a = A()
    >>> a.myaction()
    action called
    >>> is_action(a.myaction)
    True
    >>> for method in dir(A):
    ...     if is_action(getattr(A, method, None)):
    ...         print(method)
    myaction
    >>> A.myaction.__action__.name
    'my_action'
    >>> A.myaction.__action__.label
    'my action'
    """

    def __init__(  # type: ignore[explicit-any]
        self,
        name,
        label: str | None = None,
        tooltip: str | None = None,
        icon_name: str | None = None,
        shortcut: str | tuple[str, ...] | None = None,
        state: bool | Callable[[Any], bool | str] | None = None,
    ):
        self.scope: str
        self.name: str
        self.scope, self.name = name.split(".", 2) if "." in name else ("win", name)
        self.label = label
        self.tooltip = tooltip
        self.icon_name = icon_name
        self.shortcut = shortcut
        self.state = state
        self.arg_type = None
        self.task: asyncio.Task | None = None

    @property
    def detailed_name(self) -> str:
        return f"{self.scope}.{self.name}"

    @property
    def shortcuts(self) -> Iterator[str]:
        s = self.shortcut
        if isinstance(s, str):
            yield s
        elif s:
            yield from s

    def __call__(self, func):
        type_hints = get_type_hints(func)
        if "return" in type_hints:
            del type_hints["return"]
        if type_hints:
            # assume the first argument (excluding self) is our parameter
            self.arg_type = next(iter(type_hints.values()))

        if inspect.iscoroutinefunction(func):

            def async_action(*args) -> None:
                if self.task:
                    log.warning("Action %s is already running", self.detailed_name)
                    return

                coro = func(*args)
                task = asyncio.create_task(coro)
                self.task = task
                task.add_done_callback(self._task_done)

            async_action.__action__ = self  # type: ignore[attr-defined]
            return async_action

        else:
            func.__action__ = self
            return func

    def _task_done(self, task: asyncio.Task) -> None:
        self.task = None


def is_action(func):
    return hasattr(func, "__action__")



from gaphas.connector import Handle, Port

from gaphor.diagram.connectors import Connector
from gaphor.UML.actions.activity import ActivityItem, ActivityParameterNodeItem


@Connector.register(ActivityItem, ActivityParameterNodeItem)
class ActivityParameterNodeConnector:
    def __init__(
        self,
        activity: ActivityItem,
        parameter_node: ActivityParameterNodeItem,
    ) -> None:
        assert activity.diagram is parameter_node.diagram
        self.activity = activity
        self.parameter_node = parameter_node

    def allow(self, handle: Handle, port: Port) -> bool:
        return (
            bool(self.activity.diagram)
            and self.activity.diagram is self.parameter_node.diagram
            and self.parameter_node in self.activity.children
        )

    def connect(self, handle: Handle, port: Port) -> bool:
        return True

    def disconnect(self, handle: Handle) -> None:
        pass



# ruff: noqa: I001,F401,F403

# Here, order matters
import gaphor.UML.umloverrides
from gaphor.UML.uml import *
import gaphor.UML.copypaste
import gaphor.UML.deletable
import gaphor.UML.drop
import gaphor.UML.iconname
import gaphor.UML.group
import gaphor.UML.modelinglanguage
import gaphor.UML.recipes


__modeling_language__ = "UML"



import importlib
import sys

from gi.repository import Gdk, Gtk

from gaphor.abc import Service


class Styling(Service):
    def __init__(self):
        self.style_provider = init_styling()
        init_icon_theme()

    def shutdown(self):
        Gtk.StyleContext.remove_provider_for_display(
            Gdk.Display.get_default(),
            self.style_provider,
        )


def init_styling():
    style_provider = Gtk.CssProvider()
    css = (importlib.resources.files("gaphor.ui") / "styling.css").read_text(
        encoding="utf-8"
    )
    match sys.platform:
        case "win32":
            css += (
                importlib.resources.files("gaphor.ui") / "styling-windows.css"
            ).read_text(encoding="utf-8")
        case "darwin":
            css += (
                importlib.resources.files("gaphor.ui") / "styling-macos.css"
            ).read_text(encoding="utf-8")

    style_provider.load_from_string(css)

    Gtk.StyleContext.add_provider_for_display(
        Gdk.Display.get_default(),
        style_provider,
        Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
    )
    return style_provider


def init_icon_theme():
    icon_theme = Gtk.IconTheme.get_for_display(Gdk.Display.get_default())
    path = importlib.resources.files("gaphor.ui") / "icons"
    if icon_theme:
        icon_theme.add_search_path(str(path))



import getpass
import time

from gaphas.item import SE

from gaphor.core import gettext
from gaphor.diagram.diagramtoolbox import ToolDef, ToolSection, new_item_factory
from gaphor.diagram.general import diagramitems as general
from gaphor.UML import Comment, Image
from gaphor.UML.general import CommentItem, CommentLineItem, ImageItem


def metadata_config(metadata_item: general.MetadataItem) -> None:
    metadata_item.createdBy = getpass.getuser()
    metadata_item.description = metadata_item.diagram.name
    metadata_item.revision = "1.0"
    metadata_item.createdOn = time.strftime("%Y-%m-%d")


def image_config(image_item: ImageItem) -> None:
    image_item.subject.name = gettext("New Image")


general_tools = ToolSection(
    gettext("General"),
    (
        ToolDef(
            "toolbox-pointer",
            gettext("Pointer"),
            "gaphor-pointer-symbolic",
            "Escape",
            item_factory=None,
        ),
        ToolDef(
            "toolbox-magnet",
            gettext("Magnet"),
            "gaphor-magnet-symbolic",
            "F1",
            item_factory=None,
        ),
        ToolDef(
            "toolbox-line",
            gettext("Line"),
            "gaphor-line-symbolic",
            "l",
            new_item_factory(general.Line),
        ),
        ToolDef(
            "toolbox-box",
            gettext("Box"),
            "gaphor-box-symbolic",
            "b",
            new_item_factory(general.Box),
            SE,
        ),
        ToolDef(
            "toolbox-ellipse",
            gettext("Ellipse"),
            "gaphor-ellipse-symbolic",
            "e",
            new_item_factory(general.Ellipse),
            SE,
        ),
        ToolDef(
            "toolbox-comment",
            gettext("Comment"),
            "gaphor-comment-symbolic",
            "k",
            new_item_factory(CommentItem, Comment),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-comment-line",
            gettext("Comment line"),
            "gaphor-comment-line-symbolic",
            "<Shift>K",
            new_item_factory(CommentLineItem),
        ),
        ToolDef(
            "toolbox-metadata",
            gettext("Diagram metadata"),
            "gaphor-metadata-symbolic",
            None,
            new_item_factory(general.MetadataItem, config_func=metadata_config),
        ),
        ToolDef(
            "toolbox-image",
            gettext("Picture"),
            "gaphor-picture-symbolic",
            None,
            new_item_factory(ImageItem, Image, config_func=image_config),
            handle_index=SE,
        ),
    ),
)



"""State transition implementation."""

from gaphor import UML
from gaphor.diagram.presentation import LinePresentation, Named, text_name
from gaphor.diagram.shapes import Box, CssNode, Text, draw_arrow_tail
from gaphor.diagram.support import represents
from gaphor.UML.compartments import text_stereotypes


@represents(UML.Transition, head=UML.Transition.source, tail=UML.Transition.target)
class TransitionItem(Named, LinePresentation[UML.Transition]):
    """Representation of state transition."""

    def __init__(self, diagram, id=None):
        super().__init__(
            diagram,
            id,
            shape_middle=CssNode(
                "trigger-guard-action",
                None,
                Text(text=self.trigger_guard_action_text),
            ),
            shape_tail=Box(
                text_stereotypes(self),
                text_name(self),
            ),
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject[Element].appliedStereotype.classifier.name")

        self.watch("subject[Transition].trigger.name")
        self.watch("subject[Transition].guard.specification")
        self.watch("subject[Transition].action.name")

        self.draw_tail = draw_arrow_tail

    def trigger_guard_action_text(self) -> str:
        if not self.subject:
            return ""

        trigger_text = self.subject.trigger and self.subject.trigger.name or ""
        guard_text = (
            self.subject.guard
            and self.subject.guard.specification
            and f"[{UML.recipes.get_literal_value_as_string(self.subject.guard.specification)}]"
            or ""
        )
        action_text = (
            self.subject.action
            and self.subject.action.name
            and f"/{self.subject.action.name}"
            or ""
        )

        return " ".join(t for t in [trigger_text, guard_text, action_text] if t)



import logging
from pathlib import Path

from gaphor.core.modeling import Diagram
from gaphor.diagram.export import escape_filename

log = logging.getLogger(__name__)


def pkg2dir(package):
    """Return directory path from package class."""
    name: list[str] = []
    while package:
        name.insert(0, package.name)
        package = package.owningPackage
    return "/".join(name)


def export_all(factory, path, save_fn, suffix, name_re=None, underscore=None):
    for diagram in factory.select(Diagram):
        odir = f"{path}/{pkg2dir(diagram.owner)}"
        # just diagram name
        dname = escape_filename(diagram.name)
        # full diagram name including package path
        pname = f"{odir}/{dname}"

        if underscore:
            log.info("replacing underscores")
            odir = odir.replace(" ", "_")
            dname = dname.replace(" ", "_")

        if name_re and not name_re.search(pname):
            log.debug("skipping %s", pname)
            continue

        outfilename = f"{odir}/{dname}.{suffix}"

        if not Path(odir).exists():
            log.debug("creating dir %s", odir)
            Path(odir).mkdir(parents=True)

        log.info("rendering: %s -> %s...", pname, outfilename)

        save_fn(outfilename, diagram)
